/*-
 * Copyright (c) 2006 M. Warner Losh.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD: src/sys/arm/at91/at91_ssc.c,v 1.2 2006/07/14 22:30:44 imp Exp $");

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/lock.h>
#include <sys/module.h>
#include <sys/mutex.h>
#include <sys/rman.h>
#include <sys/uio.h>
#include <machine/bus.h>

#include <arm/at91/at91_sscreg.h>
#include <arm/at91/at91_pdcreg.h>

#define MAX_BUF	1024
#define NRX_BUF	2
#define ONE_BUF (MAX_BUF / NRX_BUF)

struct at91_ssc_softc
{
	device_t dev;			/* Myself */
	void *intrhand;			/* Interrupt handle */
	struct resource *irq_res;	/* IRQ resource */
	struct resource	*mem_res;	/* Memory resource */
	struct mtx sc_mtx;		/* basically a perimeter lock */
	struct cdev *cdev;
	int flags;
#define	OPENED		0x01
#define	WRITE_BUSY	0x02
	bus_dma_tag_t tag;		/* bus dma tag  */
	bus_dmamap_t tx_map;
	int txdone;
	uint8_t rx_buf[MAX_BUF];
	uint8_t *rx_ptr;
	bus_dmamap_t rx_map[NRX_BUF];
	bus_addr_t rx_pa[NRX_BUF];
	int rxcur;
	int rxdone;
	uint8_t	rd_buf[MAX_BUF];
	uint8_t *rd_end;
};

static void at91_ssc_loadread(void *arg, bus_dma_segment_t *segs, int nsegs,
    int error);
static void at91_ssc_loadwrite(void *arg, bus_dma_segment_t *segs, int nsegs,
    bus_size_t size, int error);

static inline uint32_t
RD4(struct at91_ssc_softc *sc, bus_size_t off)
{
	return bus_read_4(sc->mem_res, off);
}

static inline void
WR4(struct at91_ssc_softc *sc, bus_size_t off, uint32_t val)
{
	bus_write_4(sc->mem_res, off, val);
}

#define AT91_SSC_LOCK(_sc)		mtx_lock(&(_sc)->sc_mtx)
#define	AT91_SSC_UNLOCK(_sc)		mtx_unlock(&(_sc)->sc_mtx)
#define AT91_SSC_LOCK_INIT(_sc) \
	mtx_init(&(_sc)->sc_mtx, device_get_nameunit((_sc)->dev), \
	    "ssc", MTX_DEF)
#define AT91_SSC_LOCK_DESTROY(_sc)	mtx_destroy(&(_sc)->sc_mtx);
#define AT91_SSC_ASSERT_LOCKED(_sc)	mtx_assert(&(_sc)->sc_mtx, MA_OWNED);
#define AT91_SSC_ASSERT_UNLOCKED(_sc) mtx_assert(&(_sc)->sc_mtx, MA_NOTOWNED);
#define CDEV2SOFTC(dev)		((dev)->si_drv1)

static devclass_t at91_ssc_devclass;

/* bus entry points */

static int at91_ssc_probe(device_t dev);
static int at91_ssc_attach(device_t dev);
static int at91_ssc_detach(device_t dev);
static void at91_ssc_intr(void *);

/* helper routines */
static int at91_ssc_activate(device_t dev);
static void at91_ssc_deactivate(device_t dev);

/* cdev routines */
static d_open_t at91_ssc_open;
static d_close_t at91_ssc_close;
static d_read_t at91_ssc_read;
static d_write_t at91_ssc_write;

static struct cdevsw at91_ssc_cdevsw =
{
	.d_version = D_VERSION,
	.d_open = at91_ssc_open,
	.d_close = at91_ssc_close,
	.d_read = at91_ssc_read,
	.d_write = at91_ssc_write,
};

static int
at91_ssc_probe(device_t dev)
{
	device_set_desc(dev, "SSC");
	return (0);
}

static int
at91_ssc_attach(device_t dev)
{
	struct at91_ssc_softc *sc = device_get_softc(dev);
	int err, i;

	sc->dev = dev;
	err = at91_ssc_activate(dev);
	if (err)
		goto out;

	AT91_SSC_LOCK_INIT(sc);

	/*
	 * Activate the interrupt
	 */
	err = bus_setup_intr(dev, sc->irq_res, INTR_TYPE_MISC | INTR_MPSAFE,
	    at91_ssc_intr, sc, &sc->intrhand);
	if (err) {
		AT91_SSC_LOCK_DESTROY(sc);
		goto out;
	}

	/*
	 * Allocate DMA tags and maps
	 */
	err = bus_dma_tag_create(NULL, 1, 0, BUS_SPACE_MAXADDR_32BIT,
	    BUS_SPACE_MAXADDR, NULL, NULL, 2048, 1, 2048, BUS_DMA_ALLOCNOW,
	    busdma_lock_mutex, &sc->sc_mtx, &sc->tag);
	if (err != 0)
		goto out;
	err = bus_dmamap_create(sc->tag, 0,  &sc->tx_map);
	if (err != 0)
		goto out;
	for (i = 0; i < NRX_BUF; i++) {
		err = bus_dmamap_create(sc->tag, 0,  &sc->rx_map[i]);
		if (err != 0)
			goto out;
		err = bus_dmamap_load(sc->tag, sc->rx_map[i], sc->rx_buf +
		    ONE_BUF * i, ONE_BUF, at91_ssc_loadread, sc, 0);
		if (err != 0)
			goto out;
	}
	// We use two buffers for read, and we bounce between them.


	sc->cdev = make_dev(&at91_ssc_cdevsw, device_get_unit(dev), UID_ROOT,
	    GID_WHEEL, 0600, "ssc%d", device_get_unit(dev));
	if (sc->cdev == NULL) {
		err = ENOMEM;
		goto out;
	}
	sc->cdev->si_drv1 = sc;
	sc->rd_end = sc->rd_buf;

	// Init for TSC needs
	WR4(sc, SSC_CR, SSC_CR_SWRST);
	WR4(sc, SSC_CMR, 0);		// clock divider unused
	WR4(sc, SSC_RCMR,
	    SSC_RCMR_CKS_RK | SSC_RCMR_CKO_NONE | SSC_RCMR_START_FALL_EDGE_RF);
	WR4(sc, SSC_RFMR,
	    0x1f | SSC_RFMR_MSFBF | SSC_RFMR_FSOS_NONE);
	WR4(sc, SSC_TCMR,
	    SSC_TCMR_CKS_TK | SSC_TCMR_CKO_NONE |  SSC_TCMR_START_CONT);
	WR4(sc, SSC_TFMR,
	    0x1f | SSC_TFMR_DATDEF | SSC_TFMR_MSFBF | SSC_TFMR_FSOS_NEG_PULSE);

out:
	if (err)
		at91_ssc_deactivate(dev);
	return (err);
}

static int
at91_ssc_detach(device_t dev)
{
	return (EBUSY);	/* XXX */
}

static int
at91_ssc_activate(device_t dev)
{
	struct at91_ssc_softc *sc;
	int rid;

	sc = device_get_softc(dev);
	rid = 0;
	sc->mem_res = bus_alloc_resource_any(dev, SYS_RES_MEMORY, &rid,
	    RF_ACTIVE);
	if (sc->mem_res == NULL)
		goto errout;
	rid = 0;
	sc->irq_res = bus_alloc_resource_any(dev, SYS_RES_IRQ, &rid,
	    RF_ACTIVE);
	if (sc->irq_res == NULL)
		goto errout;
	return (0);
errout:
	at91_ssc_deactivate(dev);
	return (ENOMEM);
}

static void
at91_ssc_deactivate(device_t dev)
{
	struct at91_ssc_softc *sc;

	sc = device_get_softc(dev);
	if (sc->intrhand)
		bus_teardown_intr(dev, sc->irq_res, sc->intrhand);
	sc->intrhand = 0;
	bus_generic_detach(sc->dev);
	if (sc->mem_res)
		bus_release_resource(dev, SYS_RES_IOPORT,
		    rman_get_rid(sc->mem_res), sc->mem_res);
	sc->mem_res = 0;
	if (sc->irq_res)
		bus_release_resource(dev, SYS_RES_IRQ,
		    rman_get_rid(sc->irq_res), sc->irq_res);
	sc->irq_res = 0;
	return;
}

static void
at91_ssc_intr(void *xsc)
{
	struct at91_ssc_softc *sc = xsc;
	uint32_t status;
	int len;

	/* Reading the status also clears the interrupt */
	status = RD4(sc, SSC_SR);
	if (status == 0)
		return;
	AT91_SSC_LOCK(sc);
	if (status & SSC_SR_ENDTX) {
		WR4(sc, SSC_IDR, SSC_SR_ENDTX);
		sc->txdone++;
		wakeup(&sc->txdone);
	}
	if (status & SSC_SR_ENDRX) {
		// At this point we have at least one datum in the
		// rx_buf.  Turn off the PDC, read the read pointer,
		// move to the next set of buffers, turn on the PDC
		// and then copy the data.  The data is likely to be
		// short, so having an extra thread do this is too much
		// overhead.
		bus_dmamap_sync(sc->tag, sc->rx_map[sc->rxcur],
		    BUS_DMASYNC_POSTREAD);
		WR4(sc, PDC_PTCR, PDC_PTCR_RXTDIS);
		len = ONE_BUF - RD4(sc, PDC_RCR) * 4 + 4;
		sc->rxcur = (sc->rxcur + 1) % NRX_BUF;
		WR4(sc, PDC_RPR, sc->rx_pa[sc->rxcur]);
		WR4(sc, PDC_RCR, 1);
		WR4(sc, PDC_RNPR, sc->rx_pa[sc->rxcur] + 4);
		WR4(sc, PDC_RNCR, ONE_BUF / 4 - 1);
		bus_dmamap_sync(sc->tag, sc->rx_map[sc->rxcur],
		    BUS_DMASYNC_PREREAD);
		WR4(sc, PDC_PTCR, PDC_PTCR_RXTEN);
		if (sc->rd_end - sc->rd_buf + len < sizeof(sc->rx_buf)) {
			memcpy(sc->rd_end, sc->rx_ptr, len);
			sc->rd_end += len;
		}
		sc->rx_ptr += ONE_BUF;
		if (sc->rx_ptr >= sc->rx_buf + sizeof(sc->rx_buf))
			sc->rx_ptr = sc->rx_buf;
		sc->rxdone++;
		wakeup(&sc->rxdone);
	}
	AT91_SSC_UNLOCK(sc);
	wakeup(sc);
	return;
}

static int 
at91_ssc_open(struct cdev *dev, int oflags, int devtype, struct thread *td)
{
	struct at91_ssc_softc *sc;

	sc = CDEV2SOFTC(dev);
	AT91_SSC_LOCK(sc);
	if (!(sc->flags & OPENED)) {
		sc->flags |= OPENED;
	}
	AT91_SSC_UNLOCK(sc);
    	return (0);
}

static int
at91_ssc_close(struct cdev *dev, int fflag, int devtype, struct thread *td)
{
	struct at91_ssc_softc *sc;

	sc = CDEV2SOFTC(dev);
	AT91_SSC_LOCK(sc);
	sc->flags &= ~OPENED;
	AT91_SSC_UNLOCK(sc);
	return (0);
}

static void
at91_ssc_loadread(void *arg, bus_dma_segment_t *segs, int nsegs, int error)
{
	struct at91_ssc_softc *sc;

	sc = arg;
	if (error != 0)
		return;
	bus_dmamap_sync(sc->tag, sc->rx_map[sc->rxcur], BUS_DMASYNC_PREREAD);
	sc->rx_pa[sc->rxcur] = segs[0].ds_addr;
	if (sc->rx_ptr == NULL) {
		WR4(sc, PDC_RPR, sc->rx_pa[sc->rxcur]);
		WR4(sc, PDC_RCR, 1);
		WR4(sc, PDC_RNPR, sc->rx_pa[sc->rxcur] + 4);
		WR4(sc, PDC_RNCR, ONE_BUF / 4 - 1);
		WR4(sc, SSC_IER, SSC_SR_ENDRX);
		WR4(sc, SSC_CR, SSC_CR_RXEN);
		WR4(sc, PDC_PTCR, PDC_PTCR_RXTEN);
		sc->rx_ptr = sc->rx_buf;
	}
	sc->rxcur = (sc->rxcur + 1) % NRX_BUF;
}

static void
at91_ssc_loadwrite(void *arg, bus_dma_segment_t *segs, int nsegs,
    bus_size_t size, int error)
{
	struct at91_ssc_softc *sc;

	sc = arg;
	if (error != 0)
		return;
	bus_dmamap_sync(sc->tag, sc->tx_map, BUS_DMASYNC_PREWRITE);
	WR4(sc, PDC_TPR, segs[0].ds_addr);
	WR4(sc, PDC_TCR, size / 4);
	WR4(sc, SSC_IER, SSC_SR_ENDTX);
	WR4(sc, SSC_CR, SSC_CR_TXEN);
	WR4(sc, PDC_PTCR, PDC_PTCR_TXTEN);
}

static int
at91_ssc_read(struct cdev *dev, struct uio *uio, int flag)
{
	struct at91_ssc_softc *sc;
	int err, ret, len;

	sc = CDEV2SOFTC(dev);
	// must read a multiple of 4 bytes
	if ((uio->uio_resid & 0x3) != 0)
		return (EINVAL);
	err = 0;
	ret = 0;
	while (uio->uio_resid) {
		// If there's no buffered data, but we've already copied
		// some data, then go ahead and return what we have now.
		if (sc->rd_end == sc->rd_buf && ret != 0)
			break;
		err = msleep(&sc->rxdone, &sc->sc_mtx, PCATCH | PZERO,
		    "sscrd", 0);
		if (err != 0)
			break;
		if (sc->rd_end == sc->rd_buf)
			continue;
		len = MIN(sc->rd_end - sc->rd_buf, uio->uio_resid);
		err = uiomove(sc->rd_buf, len, uio);
		if (err != 0)
			break;
		// If we read the whole thing, no datacopy is needed,
		// otherwise we move the data down.
		ret += len;
		if (len == sc->rd_end - sc->rd_buf)
			sc->rd_end = sc->rd_buf;
		else {
			bcopy(sc->rd_buf + len, sc->rd_buf,
			    sc->rd_end - sc->rd_buf - len);
			sc->rd_end -= len;
		}
	}
	return (err);
}

static int
at91_ssc_write(struct cdev *dev, struct uio *uio, int flag)
{
	struct at91_ssc_softc *sc;
	int err, txdone;

	sc = CDEV2SOFTC(dev);
	// must write a multiple of 4 bytes
	if ((uio->uio_resid & 0x3) != 0 || uio->uio_resid > MAX_BUF)
		return (EINVAL);
	AT91_SSC_LOCK(sc);
	err = 0;
	// XXX non blocking I/O needs to bail here rather than wait.
	while (err != EINTR && sc->flags & WRITE_BUSY)
		err = msleep(&sc->txdone, &sc->sc_mtx, PCATCH | PZERO,
		    "sscwr2", 0);
	if (err == EINTR)
		return (EINTR);
	sc->flags |= WRITE_BUSY;
	WR4(sc, SSC_CR, SSC_CR_TXDIS);
	WR4(sc, PDC_PTCR, PDC_PTCR_TXTDIS);
	txdone = sc->txdone;
	err = bus_dmamap_load_uio(sc->tag, sc->tx_map, uio, at91_ssc_loadwrite,
	    sc, 0);
	if (err != 0 && err != EINPROGRESS)
		return (err);
	err = 0;
	while (txdone == sc->txdone && err != EINTR)
		err = msleep(&sc->txdone, &sc->sc_mtx, PCATCH | PZERO,
		    "sscwr", 0);
	bus_dmamap_sync(sc->tag, sc->tx_map, BUS_DMASYNC_POSTWRITE);
	sc->flags &= ~WRITE_BUSY;
	AT91_SSC_UNLOCK(sc);
	if (err == 0)
		uio->uio_resid = 0;
	return err;
}

static device_method_t at91_ssc_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,		at91_ssc_probe),
	DEVMETHOD(device_attach,	at91_ssc_attach),
	DEVMETHOD(device_detach,	at91_ssc_detach),

	{ 0, 0 }
};

static driver_t at91_ssc_driver = {
	"at91_ssc",
	at91_ssc_methods,
	sizeof(struct at91_ssc_softc),
};

DRIVER_MODULE(at91_ssc, atmelarm, at91_ssc_driver, at91_ssc_devclass, 0, 0);
