/*-
 * Copyright (c) 2006 M. Warner Losh.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_at91.h"

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/fcntl.h>
#include <sys/filio.h>
#include <sys/kernel.h>
#include <sys/lock.h>
#include <sys/module.h>
#include <sys/mutex.h>
#include <sys/poll.h>
#include <sys/rman.h>
#include <sys/selinfo.h>
#include <sys/taskqueue.h>
#include <sys/time.h>
#include <sys/timepps.h>
#include <sys/timetc.h>
#include <sys/uio.h>

#include <machine/bus.h>
#include <machine/cpu.h>
#include <machine/cpufunc.h>
#include <machine/resource.h>
#include <machine/frame.h>
#include <machine/intr.h>
#include <arm/at91/at91_tcreg.h>

struct at91_tc_softc;

#define MAX_COUNTER 3

struct at91_tc_counter {
	void *intrhand;			/* Interrupt handle */
	struct resource *irq_res;	/* IRQ resource */
	struct at91_tc_softc *sc;
	bus_size_t offset;
#ifdef AT91_TSC
	struct pps_state pps;
	struct selinfo selp;
	struct cdev *pdev;
	struct cdev *cdev;
	uint32_t pps_seen;
	uint32_t pps_read;
	struct task task;
	int	cnonblock;
#endif
};

struct at91_tc_softc {
	struct resource	*mem_res;	/* Memory resource */
	device_t dev;
	struct at91_tc_counter tc[MAX_COUNTER];
	struct mtx mtx;		/* basically a perimeter lock */
	struct cdev *cdev;
	int overflows;
};

#define RD4(tc, off) \
	bus_read_4((tc)->sc->mem_res, (off) + (tc)->offset)
#define WR4(tc, off, val) \
	bus_write_4((tc)->sc->mem_res, (off) + (tc)->offset, (val))
#define RD4sc(sc, off) \
	bus_read_4(sc->mem_res, (off))
#define WR4sc(sc, off, val) \
	bus_write_4(sc->mem_res, (off), (val))

#define AT91_TC_ASSERT_LOCKED(_sc)	mtx_assert(&_sc->mtx, MA_OWNED);
#define AT91_TC_ASSERT_UNLOCKED(_sc) mtx_assert(&_sc->mtx, MA_NOTOWNED);

#ifdef AT91_TSC
static	d_open_t	at91_tc_open;
static	d_close_t	at91_tc_close;
static	d_ioctl_t	at91_tc_ioctl;

static	d_open_t	at91_tc_ctl_open;
static	d_close_t	at91_tc_ctl_close;
static	d_read_t	at91_tc_ctl_read;
static	d_ioctl_t	at91_tc_ctl_ioctl;
static	d_poll_t	at91_tc_ctl_poll;

static task_fn_t at91_tc_tq_wakeup;

static struct cdevsw pps_cdevsw = {
	.d_version =	D_VERSION,
	.d_open =	at91_tc_open,
	.d_close =	at91_tc_close,
	.d_ioctl =	at91_tc_ioctl,
	.d_name =	"pps",
};

static struct cdevsw pps_ctl_cdevsw = {
	.d_version =	D_VERSION,
	.d_open =	at91_tc_ctl_open,
	.d_read =	at91_tc_ctl_read,
	.d_close =	at91_tc_ctl_close,
	.d_ioctl =	at91_tc_ctl_ioctl,
	.d_poll =	at91_tc_ctl_poll,
	.d_name =	"pps",
};
#endif

/* helper routines */
static int at91_tc_activate(device_t dev);
static void at91_tc_deactivate(device_t dev);
static void at91_tc_counter_init(struct at91_tc_softc *, struct at91_tc_counter *,
    int unit);

#ifdef AT91_TSC
static driver_intr_t at91_tc_counter_isr;

static struct at91_tc_counter *at91_tc_tc;

static unsigned
at91_tc_get_timecount(struct timecounter *tc)
{
	return (RD4(at91_tc_tc, TC_CV));
}

static struct timecounter at91_tc_timecounter = {
	at91_tc_get_timecount, /* get_timecount */
	NULL, /* no poll_pps */
	0xffffu, /* counter_mask */
//	5000000, /* frequency */
	60000000 / 8, /* frequency */
	"5MHz", /* name */
	2000 /* quality */
};
#endif

static int
at91_tc_probe(device_t dev)
{

	if (device_get_unit(dev) == 0)
		device_set_desc(dev, "TC0");
	else
		device_set_desc(dev, "TC3");
	return (0);
}

static int
at91_tc_attach(device_t dev)
{
	struct at91_tc_softc *sc = device_get_softc(dev);
	int err, i;

	sc->dev = dev;
	err = at91_tc_activate(dev);
	if (err)
		goto out;
	mtx_init(&sc->mtx, device_get_nameunit(sc->dev), "tc", MTX_DEF);

	sc->tc[0].offset = TC_TC0_OFFSET;
	sc->tc[1].offset = TC_TC1_OFFSET;
	sc->tc[2].offset = TC_TC2_OFFSET;

#ifdef AT91_TSC
	if (device_get_unit(dev) == 0)
		// On the TSC board, we have 5MHz going into TIOA2.  Setup
		// TC1's XC1 to use this.
		WR4sc(sc, TC_BMR, TC_BMR_XC1_TIOA2);
#endif
	for (i = 0; i < MAX_COUNTER; i++)
		at91_tc_counter_init(sc, &sc->tc[i], i);
#ifdef AT91_TSC
	if (device_get_unit(dev) == 0) {
		// XXX need cdev
		// Init our timecounter.
		at91_tc_tc = &sc->tc[1];
		tc_init(&at91_tc_timecounter);
	}
#endif
out:;
	return (err);
}

static int
at91_tc_activate(device_t dev)
{
	struct at91_tc_softc *sc;
	int rid;

	sc = device_get_softc(dev);
	rid = 0;
	sc->mem_res = bus_alloc_resource_any(dev, SYS_RES_MEMORY, &rid,
	    RF_ACTIVE);
	if (sc->mem_res == NULL)
		goto errout;
	return (0);
errout:;
	at91_tc_deactivate(dev);
	return (ENOMEM);
}

static void
at91_tc_tc_deactivate(struct at91_tc_counter *tc)
{
	if (tc->intrhand)
		bus_teardown_intr(tc->sc->dev, tc->irq_res, tc->intrhand);
	tc->intrhand = 0;
	if (tc->irq_res)
		bus_release_resource(tc->sc->dev, SYS_RES_IRQ,
		    rman_get_rid(tc->irq_res), tc->irq_res);
	tc->irq_res = 0;
}

static void
at91_tc_deactivate(device_t dev)
{
	struct at91_tc_softc *sc;
	int i;

	sc = device_get_softc(dev);
	for (i = 0; i < MAX_COUNTER; i++)
		at91_tc_tc_deactivate(&sc->tc[i]);
	bus_generic_detach(sc->dev);
	if (sc->mem_res)
		bus_release_resource(dev, SYS_RES_IOPORT,
		    rman_get_rid(sc->mem_res), sc->mem_res);
	sc->mem_res = 0;
	return;
}

static void
at91_tc_counter_init(struct at91_tc_softc *sc, struct at91_tc_counter *tc,
    int unit)
{
#ifdef AT91_TSC
	int rid;
	struct cdev *d;
#endif

	tc->sc = sc;
#ifdef AT91_TSC
	// Only TC1 is needed.  All others are disabled.
	if (unit != 1 || device_get_unit(sc->dev) != 0) {
#endif
		WR4(tc, TC_IDR, 0xffffffff);
		WR4(tc, TC_CCR, TC_CCR_CLKDIS);
		return;
#ifdef AT91_TSC
	}

	rid = unit;
	tc->irq_res = bus_alloc_resource_any(sc->dev, SYS_RES_IRQ, &rid,
	    RF_ACTIVE);
	if (tc->irq_res == NULL)
		return;
	if (bus_setup_intr(sc->dev, tc->irq_res, INTR_TYPE_MISC | INTR_FAST,
	    at91_tc_counter_isr, tc, &tc->intrhand) != 0)
		return;

	TASK_INIT(&tc->task, 0, at91_tc_tq_wakeup, tc);
	rid = 0;
	d = make_dev(&pps_cdevsw, unit, UID_ROOT, GID_WHEEL, 0600,
	    "pps%d", rid);
	d->si_drv1 = tc;
	d->si_drv2 = (void *)0;
	tc->pdev = d;
	tc->pps.ppscap = PPS_CAPTUREASSERT;
	pps_init(&tc->pps);
	d = make_dev(&pps_ctl_cdevsw, unit, UID_ROOT, GID_WHEEL, 0600,
	    "pps%d.ctl", rid);
	d->si_drv1 = tc;
	d->si_drv2 = (void *)1;
	tc->cdev = d;

	// Setup TC1 into Capture Mode (WAVE=0), clocked by our external
	// 5MHz, loading RA on rise and RB on falling edge.  For some reason,
	// we seem to need to enable both of these captures and throw away
	// the one we don't need, otherwise neither happens.
	WR4(tc, TC_CMR,
#if 0
	    TC_CMR_XC1 |
#else
	    TC_CMR_TIMER_CLOCK2 |
#endif
	    TC_CMR_BURST_NONE | TC_CMR_ETRGEDG_NONE |
	    TC_CMR_LDRA_RISE | TC_CMR_LDRB_FALL);
	WR4(tc, TC_IDR, 0xffffffff);
	WR4(tc, TC_IER, TC_SR_COVFS | TC_SR_LOVRS | TC_SR_LDRAS | TC_SR_LDRBS);

	// Now that we have ISR setup, we can enable clocks and go!
	WR4(tc, TC_CCR, TC_CCR_SWTRG | TC_CCR_CLKEN);
#endif
}

#ifdef AT91_TSC
static	int
at91_tc_open(struct cdev *dev, int flags, int fmt, struct thread *td)
{
	return(0);
}

static	int
at91_tc_close(struct cdev *dev, int flags, int fmt, struct thread *td)
{
	struct at91_tc_counter *tc = dev->si_drv1;

	mtx_lock(&tc->sc->mtx);
	tc->pps.ppsparam.mode = 0;	/* PHK ??? */
	mtx_unlock(&tc->sc->mtx);
	return(0);
}

static int
at91_tc_ioctl(struct cdev *dev, u_long cmd, caddr_t data, int flags,
    struct thread *td)
{
	struct at91_tc_counter *tc = dev->si_drv1;
	int err;

	mtx_lock(&tc->sc->mtx);
	err = pps_ioctl(cmd, data, &tc->pps);
	mtx_unlock(&tc->sc->mtx);
	return (err);
}

static	int
at91_tc_ctl_open(struct cdev *dev, int flags, int fmt, struct thread *td)
{
	struct at91_tc_counter *tc = dev->si_drv1;
	mtx_lock(&tc->sc->mtx);
	if (flags & O_NONBLOCK)
		tc->cnonblock = 1;
	else
		tc->cnonblock = 0;
	tc->pps_read = tc->pps_seen;
	mtx_unlock(&tc->sc->mtx);
	return(0);
}

static	int
at91_tc_ctl_close(struct cdev *dev, int flags, int fmt, struct thread *td)
{
	return(0);
}

static int
at91_tc_ctl_poll(struct cdev *dev, int events, struct thread *td)
{
	struct at91_tc_counter *tc = dev->si_drv1;
	int revents = 0;

	mtx_lock(&tc->sc->mtx);
	if (events & (POLLIN | POLLRDNORM)) {
		if (tc->pps_seen != tc->pps_read)
			revents |= events & (POLLIN | POLLRDNORM);
		else
			selrecord(td, &tc->selp);
	}
	mtx_unlock(&tc->sc->mtx);

	return (revents);
}

static int
at91_tc_ctl_ioctl(struct cdev *dev, u_long cmd, caddr_t arg, int flags,
    struct thread *td)
{
	struct at91_tc_counter *tc = dev->si_drv1;

	mtx_lock(&tc->sc->mtx);
	switch (cmd) {
	case FIONBIO:
		if (*(int*)arg)
			tc->cnonblock = 1;
		else
			tc->cnonblock = 0;
		break;

#if 0
	case TSCPPSIOC_FLUSH:
		tc->pps_read = tc->pps_seen;
		break;

	case TSCPPSIOC_GET_COUNT:
		*(int *)arg = tc->pps_seen;
		break;
#endif

	default:
		mtx_unlock(&tc->sc->mtx);
		return (ENOTTY);
	}
	mtx_unlock(&tc->sc->mtx);
	return (0);
}

static int
at91_tc_ctl_read(struct cdev *dev, struct uio *uio, int ioflag)
{
	struct at91_tc_counter *tc = dev->si_drv1;
	uint8_t val;
	int err = 0;

	mtx_lock(&tc->sc->mtx);
	while (err == 0) {
		if (tc->pps_seen != tc->pps_read) {
			tc->pps_read++;
			mtx_unlock(&tc->sc->mtx);
			val = 'p';
			return (uiomove(&val, 1, uio));
		}
		if (tc->cnonblock) {
			mtx_unlock(&tc->sc->mtx);
			return EWOULDBLOCK;
		}
		err = msleep(tc, &tc->sc->mtx, PZERO, "ctlread", 0);
	}
	mtx_unlock(&tc->sc->mtx);
	return (err);
}

static void
at91_tc_tq_wakeup(void *arg, int pending)
{
	struct at91_tc_counter *tc = (struct at91_tc_counter *)arg;

	/*
	 * pps_even modifies ppsinfo.  This is also accessed and modified
	 * by the ioctl routines, so we need an interlock.  We use the same
	 * interlock to proect pps_seen.  The read code interlocks against
	 * this.  Given the low data rate for this driver (pps), contention
	 * is not a consideration.
	 */
	mtx_lock(&tc->sc->mtx);
	pps_event(&tc->pps, PPS_CAPTUREASSERT);

	/*
	 * create a PPS event so that the ctl driver can return it via the
	 * read channel.
	 */
	tc->pps_seen++;
	mtx_unlock(&tc->sc->mtx);
	selwakeup(&tc->selp);
	wakeup(tc);
}

static void
at91_tc_counter_isr(void *argp)
{
	struct at91_tc_counter *tc = argp;
	uint32_t status, r;

	status = RD4(tc, TC_SR);
	if (status & TC_SR_COVFS)
		tc->sc->overflows++;
	if (status & TC_SR_LOVRS)
		printf("Didn't read RA or RB in time\n");
	if (status & TC_SR_LDRAS) {
		r = RD4(tc, TC_RA) & 0xffff;
		pps_capture(&tc->pps);
		tc->pps.capcount = r;
		taskqueue_enqueue_fast(taskqueue_fast, &tc->task);
	}
	if (status & TC_SR_LDRBS)
		RD4(tc, TC_RB);
}
#endif

static device_method_t at91_tc_methods[] = {
	DEVMETHOD(device_probe, at91_tc_probe),
	DEVMETHOD(device_attach, at91_tc_attach),
	{0, 0},
};

static driver_t at91_tc_driver = {
	"at91_tc",
	at91_tc_methods,
	sizeof(struct at91_tc_softc),
};
static devclass_t at91_tc_devclass;

DRIVER_MODULE(at91_tc, atmelarm, at91_tc_driver, at91_tc_devclass, 0, 0);
