/* $NetBSD: ep93xx_intr.c,v 1.3 2005/01/05 04:53:50 joff Exp $ */

/*
 * Copyright (c) 2002 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Jesse Off
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Ichiro FUKUHARA and Naoto Shimazaki.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * $NetBSD: ep93xx_intr.c,v 1.3 2005/01/05 04:53:50 joff Exp $
 *
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD");

/*
 * Interrupt support for the Cirrus Logic EP93XX
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/bus.h>
#define _ARM32_BUS_DMA_PRIVATE
#define __RMAN_RESOURCE_VISIBLE
#include <sys/rman.h>

#include <machine/bus.h>
#include <machine/intr.h>

#include <machine/cpufunc.h>

#include <arm/ep93xx/ep93xxreg.h> 
#include <arm/ep93xx/ep93xxvar.h> 
#include <arm/ep93xx/ep93xx.h>

#include <arm/ep93xx/tsarmreg.h> /* XXX */

MALLOC_DEFINE(M_AHPBDEV, "AHB & APB Device", "");

struct ahpb_device {
	struct resource_list	ad_resources;
};

struct ahb_softc {
	struct rman	sc_intr_rman;
	struct rman	sc_mem_rman;
};

struct apb_softc {
	struct rman	sc_mem_rman;
};

static device_t
ahpb_add_child(device_t dev, const char *name, int unit)
{
        device_t child;
        struct  ahpb_device *adev;
	int	start, count;

        child = device_add_child(dev, name, unit);
        if (child == NULL) 
                return (child);
        
        adev = malloc(sizeof(struct ahpb_device), M_AHPBDEV, M_NOWAIT | M_ZERO);
        if (adev == NULL)
                return (NULL);

        resource_list_init(&adev->ad_resources);

        device_set_ivars(child, adev);

        start = 0;
        count = 0;
        resource_int_value(name, unit, "port", &start);
        resource_int_value(name, unit, "portsize", &count);
        if (start > 0 || count > 0)
                bus_set_resource(child, SYS_RES_MEMORY, -1, start, count);

	if (resource_int_value(name, unit, "irq", &start) == 0 && start > 0)
		bus_set_resource(child, SYS_RES_IRQ, -1, start, 1);

	if (resource_int_value(name, unit, "ppalloc", &start) == 0 && start > 0)
		bus_set_resource(child, SYS_RES_DRQ, -1, start, 1);

        return (child);
}

static void
ahpb_add_hint_devs(device_t bus)
{
	char buf[16];
	const char *dname;
	int dunit;
	int i;

	/*
	 * Add all devices configured to be attached to parent.
	 */
	snprintf(buf, sizeof(buf), "%s", device_get_name(bus));
	i = 0;
	while ((resource_find_match(&i, &dname, &dunit, "at", buf)) == 0) {
		if (bootverbose)
			printf("adding %s%d on %s from hint\n", dname, dunit,
			    buf);
		ahpb_add_child(bus, dname, dunit);
	}


}

static struct resource_list *
ahpb_get_resource_list(device_t dev, device_t child)
{
	struct ahpb_device *adev;
	struct resource_list *rl;

	adev = (struct ahpb_device *)device_get_ivars(child);
	rl = &adev->ad_resources;

	return (rl);
}

static int
ahpb_set_resource(device_t dev, device_t child, int type, int rid,
	u_long start, u_long count)
{
	struct ahpb_device *adev;
	struct resource_list *rl;

	adev = (struct ahpb_device *)device_get_ivars(child);
	rl = &adev->ad_resources;

	if (type != SYS_RES_MEMORY && type != SYS_RES_IRQ &&
	    type != SYS_RES_DRQ)
		return (EINVAL);
	if (rid < -1)
		return (EINVAL);

	if (rid == -1)
		resource_list_add_next(rl, type, start, start + count - 1,
		    count);
	else
		resource_list_add(rl, type, rid, start, start + count - 1,
		    count);

	return 0;
}

int	ahb_probe(device_t);
void	ahb_identify(driver_t *, device_t);
int	ahb_attach(device_t);

int
ahb_probe(device_t dev)
{
	device_set_desc(dev, "Cirrus Logic EP93xx AHB Bus");
	return (0);
}

void
ahb_identify(driver_t *driver, device_t parent)
{
	
	BUS_ADD_CHILD(parent, 0, "ahb", 0);
}

struct arm32_dma_range *
bus_dma_get_range(void)
{

	return (NULL);
}

int
bus_dma_get_range_nb(void)
{

	return (0);
}

int
ahb_attach(device_t dev)
{
	struct ahb_softc *sc;

	sc = device_get_softc(dev);
	sc->sc_intr_rman.rm_type = RMAN_ARRAY;
	sc->sc_intr_rman.rm_descr = "Interrupts";
	sc->sc_mem_rman.rm_type = RMAN_ARRAY;
	sc->sc_mem_rman.rm_descr = "Register Memory";
	if (rman_init(&sc->sc_intr_rman) != 0 ||
	    rman_init(&sc->sc_mem_rman) != 0 ||
	    rman_manage_region(&sc->sc_intr_rman, 0, NIRQ - 1) != 0 ||
	    rman_manage_region(&sc->sc_mem_rman, 0, EP93XX_AHB_SIZE - 1) != 0)
		panic("ahb_attach(): failed to set up rmans");

	ep93xx_intr_init();

	/* add all the deivces per hints */
	ahpb_add_hint_devs(dev);

	/* kick these off */
	bus_generic_probe(dev);
	bus_generic_attach(dev);

	return (0);
}

void
cpu_reset(void)
{
	/*
	 * Make really really sure that all interrupts are disabled,
	 * and poke the Internal Bus and Peripheral Bus reset lines.
	 */
	(void) disable_interrupts(I32_bit|F32_bit);
	{
	 	u_int32_t feed, ctrl;
		
		feed = TS7XXX_IO16_VBASE + TS7XXX_WDOGFEED;
		ctrl = TS7XXX_IO16_VBASE + TS7XXX_WDOGCTRL;
		
		__asm __volatile (
	    	    "mov r0, #0x5\n"
		    "mov r1, #0x1\n"
    		    "strh r0, [%0]\n"
		    "strh r1, [%1]\n"
		    :
		    : "r" (feed), "r" (ctrl)
		    : "r0", "r1"
    		    );
    	}
	while (1);
}

static struct resource *
ahb_alloc_resource(device_t bus, device_t child, int type, int *rid,
    u_long start, u_long end, u_long count, u_int flags)
{
	struct ahb_softc *sc = device_get_softc(bus);
	struct resource *rv;
	struct rman *rm;
	u_long rstart, rcount;
	int error;
	int needactivate = flags & RF_ACTIVE;

	flags &= ~RF_ACTIVE;

	switch (type) {
	case SYS_RES_IRQ:
		rm = &sc->sc_intr_rman;
		break;
	case SYS_RES_MEMORY:
		rm = &sc->sc_mem_rman;
		break;
	default:
		return (NULL);
	}

	error = BUS_GET_RESOURCE(bus, child, type, *rid, &rstart, &rcount);
	if (error)
		return (NULL);

	rv = rman_reserve_resource(rm, rstart, rstart + rcount - 1, rcount,
	    flags, child);

	if (rv == NULL)
		return (NULL);
	if (type == SYS_RES_MEMORY) {
		rman_set_bustag(rv, &ep93xx_bs_tag);
		rman_set_bushandle(rv, EP93XX_AHB_VBASE + rman_get_start(rv));
	}

	if (needactivate) {
		if (bus_activate_resource(child, type, *rid, rv) != 0) {
			rman_release_resource(rv);
			return (NULL);
		}
	}

	return (rv);
}

static int
ahb_activate_resource(device_t dev, device_t child, int type, int rid,
    struct resource *r)
{
	/* XXX - enable interrupt? */
	/* not much to do since we use static mappings */
	return rman_activate_resource(r);
}

static int
ahb_deactivate_resource(device_t dev, device_t child, int type, int rid,
    struct resource *r)
{
	/* XXX - disable interrupt? */
	/* not much to do since we use static mappings */
	return rman_deactivate_resource(r);
}

static int
ahb_release_resource(device_t dev, device_t child, int type, int rid,
    struct resource *r)
{
	/* not much to do since we use static mappings */
	return rman_release_resource(r);
}

static int
ahb_setup_intr(device_t dev, device_t child, struct resource *ires,
	int flags, driver_intr_t *intr, void *arg, void **cookiep)
{

	BUS_SETUP_INTR(device_get_parent(dev), child, ires, flags, intr, arg,
	    cookiep);
	arm_unmask_irq(ires->r_start);
	
	return (0);
}

static int ahb_teardown_intr(device_t dev, device_t child,
	struct resource *ires, void *cookie)
{

	BUS_TEARDOWN_INTR(device_get_parent(dev), child, ires, cookie);
	arm_mask_irq(ires->r_start);
	
	return (0);
}

static device_method_t ahb_methods[] = {
	DEVMETHOD(device_probe,			ahb_probe),
	DEVMETHOD(device_attach,		ahb_attach),
	DEVMETHOD(device_identify,		ahb_identify),

	DEVMETHOD(bus_get_resource_list,	ahpb_get_resource_list),
	DEVMETHOD(bus_alloc_resource,		ahb_alloc_resource),
	DEVMETHOD(bus_release_resource,		ahb_release_resource),
	DEVMETHOD(bus_set_resource,		ahpb_set_resource),
	DEVMETHOD(bus_get_resource,		bus_generic_rl_get_resource),
	DEVMETHOD(bus_delete_resource,		bus_generic_rl_delete_resource),
	DEVMETHOD(bus_activate_resource,	ahb_activate_resource),
	DEVMETHOD(bus_deactivate_resource,	ahb_deactivate_resource),

	DEVMETHOD(bus_setup_intr,		ahb_setup_intr),
	DEVMETHOD(bus_teardown_intr,		ahb_teardown_intr),
	{0, 0},
};

static driver_t ahb_driver = {
	"ahb",
	ahb_methods,
	sizeof(struct ahb_softc),
};
static devclass_t ahb_devclass;

DRIVER_MODULE(ahb, nexus, ahb_driver, ahb_devclass, 0, 0);

static int
apb_probe(device_t dev)
{

	return 0;
}

static int
apb_attach(device_t dev)
{
	struct apb_softc *sc;

	sc = device_get_softc(dev);
	sc->sc_mem_rman.rm_type = RMAN_ARRAY;
	sc->sc_mem_rman.rm_descr = "Register Memory";
	if (rman_init(&sc->sc_mem_rman) != 0 ||
	    rman_manage_region(&sc->sc_mem_rman, 0, EP93XX_APB_SIZE - 1) != 0)
		panic("apb_attach(): failed to set up rmans");

	/* add all the deivces per hints */
	ahpb_add_hint_devs(dev);

	/* kick these off */
	bus_generic_probe(dev);
	bus_generic_attach(dev);

	return (0);
}

static struct resource *
apb_alloc_resource(device_t bus, device_t child, int type, int *rid,
    u_long start, u_long end, u_long count, u_int flags)
{
	struct apb_softc *sc = device_get_softc(bus);
	struct resource *rv;
	struct rman *rm;
	u_long rstart, rcount;
	int error;
	int needactivate = flags & RF_ACTIVE;

	if (type == SYS_RES_IRQ) {
		return BUS_ALLOC_RESOURCE(device_get_parent(bus), child, type,
		    rid, start, end, count, flags);
	}

	flags &= ~RF_ACTIVE;

	if (type != SYS_RES_MEMORY)
		return NULL;

	rm = &sc->sc_mem_rman;

	error = BUS_GET_RESOURCE(bus, child, type, *rid, &rstart, &rcount);
	if (error)
		return (NULL);

	rv = rman_reserve_resource(rm, rstart, rstart + rcount - 1, rcount,
	    flags, child);

	if (rv == NULL)
		return (NULL);
	if (type == SYS_RES_MEMORY) {
		/* XXX - should base this off of AHB's tag and handle */
		rman_set_bustag(rv, &ep93xx_bs_tag);
		rman_set_bushandle(rv, EP93XX_APB_VBASE + rman_get_start(rv));
	}

	if (needactivate) {
		if (bus_activate_resource(child, type, *rid, rv) != 0) {
			rman_release_resource(rv);
			return (NULL);
		}
	}

	return (rv);
}

static int
apb_activate_resource(device_t dev, device_t child, int type, int rid,
    struct resource *r)
{
	/* XXX - enable interrupt? */
	/* not much to do since we use static mappings */
	return rman_activate_resource(r);
}

static int
apb_deactivate_resource(device_t dev, device_t child, int type, int rid,
    struct resource *r)
{
	/* XXX - disable interrupt? */
	/* not much to do since we use static mappings */
	return rman_deactivate_resource(r);
}

static int
apb_release_resource(device_t dev, device_t child, int type, int rid,
    struct resource *r)
{
	/* not much to do since we use static mappings */
	return rman_release_resource(r);
}

static device_method_t apb_methods[] = {
	DEVMETHOD(device_probe,			apb_probe),
	DEVMETHOD(device_attach,		apb_attach),

	DEVMETHOD(bus_get_resource_list,	ahpb_get_resource_list),
	DEVMETHOD(bus_alloc_resource,		apb_alloc_resource),
	DEVMETHOD(bus_release_resource,		apb_release_resource),
	DEVMETHOD(bus_set_resource,		ahpb_set_resource),
	DEVMETHOD(bus_get_resource,		bus_generic_rl_get_resource),
	DEVMETHOD(bus_delete_resource,		bus_generic_rl_delete_resource),
	DEVMETHOD(bus_activate_resource,	apb_activate_resource),
	DEVMETHOD(bus_deactivate_resource,	apb_deactivate_resource),

	DEVMETHOD(bus_setup_intr,		bus_generic_setup_intr),
	DEVMETHOD(bus_teardown_intr,		bus_generic_teardown_intr),
	{0, 0},
};

static driver_t apb_driver = {
	"apb",
	apb_methods,
	sizeof(struct apb_softc),
};
static devclass_t apb_devclass;

DRIVER_MODULE(apb, ahb, apb_driver, apb_devclass, 0, 0);
