/*
 * Copyright (c) 2003 Marcel Moolenaar
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/cons.h>
#include <sys/tty.h>
#include <machine/bus.h>

#include <dev/uart/uart.h>
#include <dev/uart/uart_cpu.h>
#include <dev/uart/uart_bus.h>
#include <arm/s3c2xx0/uart_dev_s3c2410.h>
#include <arm/s3c2xx0/s3c2xx0reg.h>
#include "uart_if.h"

#define      DEFAULT_RCLK    3686400

extern int got_mmu;

/*
 * Low-level UART interface.
 */
static int s3c2410_probe(struct uart_bas *bas);
static void s3c2410_init(struct uart_bas *bas, int, int, int, int);
static void s3c2410_term(struct uart_bas *bas);
static void s3c2410_putc(struct uart_bas *bas, int);
static int s3c2410_poll(struct uart_bas *bas);
static int s3c2410_getc(struct uart_bas *bas, struct mtx *mtx);

int did_mmu = 0;

extern SLIST_HEAD(uart_devinfo_list, uart_devinfo) uart_sysdevs;

struct uart_ops uart_s3c2410_ops = {
	.probe = s3c2410_probe,
	.init = s3c2410_init,
	.term = s3c2410_term,
	.putc = s3c2410_putc,
	.poll = s3c2410_poll,
	.getc = s3c2410_getc,
};

static int
s3c2410_probe(struct uart_bas *bas)
{
	return (0);
}

static void
s3c2410_addr_change(struct uart_bas *bas)
{
	
	bas->bsh = 0xd000d000;
	did_mmu = 1;
}

static int
sscomspeed(long speed, long frequency)
{
#define	divrnd(n, q)	(((n)*2/(q)+1)/2)	/* divide and round off */

	int x, err;

	if (speed <= 0)
		return -1;
	x = divrnd(frequency / 16, speed);
	if (x <= 0)
		return -1;
	err = divrnd(((quad_t)frequency) * 1000 / 16, speed * x) - 1000;
	if (err < 0)
		err = -err;
	return x-1;

#undef	divrnd
}
static void
s3c2410_init(struct uart_bas *bas, int baudrate, int databits, int stopbits,
    int parity)
{
	int brd;
	
	/* XXX: sigh. */
	if (!did_mmu && got_mmu) 
		s3c2410_addr_change(bas);
	if (bas->rclk == 0)
		bas->rclk = DEFAULT_RCLK;
	uart_setreg(bas, SSCOM_ULCON, 0x23);
	uart_setreg(bas, SSCOM_UCON, 0);
	uart_setreg(bas, SSCOM_UFCON,
	    UFCON_TXTRIGGER_8 | UFCON_RXTRIGGER_8 |
	    UFCON_TXFIFO_RESET | UFCON_RXFIFO_RESET |
	    UFCON_FIFO_ENABLE);
	brd = sscomspeed(baudrate, bas->rclk);
	uart_setreg(bas, SSCOM_UBRDIV, brd);
	/* Enable UART. */
	uart_setreg(bas, SSCOM_UCON, UCON_TXMODE_INT|UCON_RXMODE_INT);
	uart_setreg(bas, SSCOM_UMCON, UMCON_RTS);
}

static void
s3c2410_term(struct uart_bas *bas)
{
	/* XXX */
}

static void
s3c2410_putc(struct uart_bas *bas, int c)
{
	/* XXX: sigh. */
	if (!did_mmu && got_mmu) 
		s3c2410_addr_change(bas);

#if 0
	while (uart_getreg(bas, SSCOM_UFSTAT) & UFSTAT_TXFULL);
#endif
	uart_setreg(bas, SSCOM_UTXH, c);
}

static int
s3c2410_poll(struct uart_bas *bas)
{
	return (sscom_getc(bas->bst, bas->bsh));
}

static int
s3c2410_getc(struct uart_bas *bas, struct mtx *mtx)
{
	int c;
	/* XXX: sigh. */
	if (!did_mmu && got_mmu) 
		s3c2410_addr_change(bas);

	while (!sscom_rxrdy(bas->bst, bas->bsh));
	return sscom_getc(bas->bst, bas->bsh);
	return (c);
}

static int s3c2410_bus_probe(struct uart_softc *sc);
static int s3c2410_bus_attach(struct uart_softc *sc);
static int s3c2410_bus_flush(struct uart_softc *, int);
static int s3c2410_bus_getsig(struct uart_softc *);
static int s3c2410_bus_ioctl(struct uart_softc *, int, intptr_t);
static int s3c2410_bus_ipend(struct uart_softc *);
static int s3c2410_bus_param(struct uart_softc *, int, int, int, int);
static int s3c2410_bus_receive(struct uart_softc *);
static int s3c2410_bus_setsig(struct uart_softc *, int);
static int s3c2410_bus_transmit(struct uart_softc *);

static kobj_method_t s3c2410_methods[] = {
	KOBJMETHOD(uart_probe,		s3c2410_bus_probe),
	KOBJMETHOD(uart_attach, 	s3c2410_bus_attach),
	KOBJMETHOD(uart_flush,		s3c2410_bus_flush),
	KOBJMETHOD(uart_getsig,		s3c2410_bus_getsig),
	KOBJMETHOD(uart_ioctl,		s3c2410_bus_ioctl),
	KOBJMETHOD(uart_ipend,		s3c2410_bus_ipend),
	KOBJMETHOD(uart_param,		s3c2410_bus_param),
	KOBJMETHOD(uart_receive,	s3c2410_bus_receive),
	KOBJMETHOD(uart_setsig,		s3c2410_bus_setsig),
	KOBJMETHOD(uart_transmit,	s3c2410_bus_transmit),
	
	{0, 0 }
};

int
s3c2410_bus_probe(struct uart_softc *sc)
{
	return (0);
}

static int
s3c2410_bus_attach(struct uart_softc *sc)
{
	 bcopy(&sc->sc_sysdev->bas, &sc->sc_bas, sizeof(sc->sc_bas));

	 sc->sc_txfifosz = 3;
	 sc->sc_rxfifosz = 1;
	 sc->sc_hwiflow = 0;
	return (0);
}
static int
s3c2410_bus_transmit(struct uart_softc *sc)
{

	return (0);
}
static int
s3c2410_bus_setsig(struct uart_softc *sc, int sig)
{
	return (0);
}
static int
s3c2410_bus_receive(struct uart_softc *sc)
{
	
	return (0);
}
static int
s3c2410_bus_param(struct uart_softc *sc, int baudrate, int databits,
    int stopbits, int parity)
{
	
	return (0);
}
static int
s3c2410_bus_ipend(struct uart_softc *sc)
{
	return (0);
}
static int
s3c2410_bus_flush(struct uart_softc *sc, int what)
{
	return (0);
}

static int
s3c2410_bus_getsig(struct uart_softc *sc)
{
	return (0);
}

static int
s3c2410_bus_ioctl(struct uart_softc *sc, int request, intptr_t data)
{
	return (EINVAL);
}
struct uart_class uart_s3c2410_class = {
	"s3c2410 class",
	s3c2410_methods,
	1,
	.uc_range = 8,
	.uc_rclk = 3686400
};
