/*******************************************************************************
 *
 * Filename: eeprom.c
 *
 * Instantiation of eeprom routines
 *
 * Revision information:
 *
 * 28AUG2004	kb_admin	initial creation - adapted from Atmel sources
 * 12JAN2005	kb_admin	fixed clock generation, write polling
 *
 * BEGIN_KBDD_BLOCK
 * No warranty, expressed or implied, is included with this software.  It is
 * provided "AS IS" and no warranty of any kind including statutory or aspects
 * relating to merchantability or fitness for any purpose is provided.  All
 * intellectual property rights of others is maintained with the respective
 * owners.  This software is not copyrighted and is intended for reference
 * only.
 * END_BLOCK
 ******************************************************************************/

#include "at91rm9200_lowlevel.h"
#include "eeprom.h"
#include "AT91RM9200.h"
#include "prompt.h"

/* ****************************** GLOBALS *************************************/


/* ********************** PRIVATE FUNCTIONS/DATA ******************************/


/* Use a macro to calculate the TWI clock generator value to save code space. */
#define TWI_CLK_BASE_DIV	((AT91C_MASTER_CLOCK/(4*AT91C_TWSI_CLOCK)) - 2)
#define SET_TWI_CLOCK	((0x00010000) | (TWI_CLK_BASE_DIV) | (TWI_CLK_BASE_DIV << 8))


parse_function_t	eeread_function, eewrite_function;

/* ************************** GLOBAL FUNCTIONS ********************************/


/*
 * .KB_C_FN_DEFINITION_START
 * void ReadEEPROM(unsigned ee_addr, char *data_addr, unsigned size)
 *  This global function reads data from the eeprom at ee_addr storing data
 * to data_addr for size bytes.  Assume the TWI has been initialized.
 * This function does not utilize the page read mode to simplify the code.
 * .KB_C_FN_DEFINITION_END
 */
void
ReadEEPROM(unsigned ee_addr, char *data_addr, unsigned size)
{
	const AT91PS_TWI 	twiPtr = AT91C_BASE_TWI;
	unsigned int status;

	status = twiPtr->TWI_SR;
	status = twiPtr->TWI_RHR;

	// Set the TWI Master Mode Register
	twiPtr->TWI_MMR = (TWSI_EEPROM_ADDRESS << 16) | AT91C_TWI_IADRSZ_2_BYTE | AT91C_TWI_MREAD;

	// Set TWI Internal Address Register
	twiPtr->TWI_IADR = ee_addr;

	// Start transfer
	twiPtr->TWI_CR = AT91C_TWI_START;

	status = twiPtr->TWI_SR;

	while (size-- > 1){

		// Wait RHR Holding register is full
		while (!(twiPtr->TWI_SR & AT91C_TWI_RXRDY));

		// Read byte
		*(data_addr++) = twiPtr->TWI_RHR;
	}

	twiPtr->TWI_CR = AT91C_TWI_STOP;

	status = twiPtr->TWI_SR;

	// Wait transfer is finished
	while (!(twiPtr->TWI_SR & AT91C_TWI_TXCOMP));

	// Read last byte
	*data_addr = twiPtr->TWI_RHR;
}


/*
 * .KB_C_FN_DEFINITION_START
 * int eeread_parse(int argc, char *argv[])
 *  This global function parses text from the command line.
 * .KB_C_FN_DEFINITION_END
 */
int eeread_parse(int argc, char *argv[]) {

	unsigned eeAddr, srcAddr, size;

	if (argc < 4) {
		DebugPrint("Missing parameter");
		return (1);
	}

	p_ASCIIToHex(argv[1], &eeAddr);
	p_ASCIIToHex(argv[2], &srcAddr);
	p_ASCIIToHex(argv[3], &size);
	ReadEEPROM(eeAddr, (char*)srcAddr, size);

	return (0);
}


/*
 * .KB_C_FN_DEFINITION_START
 * void eeread_help(void)
 *  This global displays help info for the corresponding command.
 * .KB_C_FN_DEFINITION_END
 */
void eeread_help(void) {
	DebugPrint("\tread EEPROM to memory range\n\r");
	DebugPrint("\t\teeread <ee_addr> <dst_data_addr> <size>\n\r");
}


/*
 * .KB_C_FN_DEFINITION_START
 * void WriteEEPROM(unsigned ee_addr, char *data_addr, unsigned size)
 *  This global function writes data to the eeprom at ee_addr using data
 * from data_addr for size bytes.  Assume the TWI has been initialized.
 * This function does not utilize the page write mode as the write time is
 * much greater than the time required to access the device for byte-write
 * functionality.  This allows the function to be much simpler.
 * .KB_C_FN_DEFINITION_END
 */
void WriteEEPROM(unsigned ee_addr, char *data_addr, unsigned size) {

	const AT91PS_TWI 	twiPtr = AT91C_BASE_TWI;
	unsigned		status;
	char			test_data;

	while (size--) {
		if (!(ee_addr & 0x3f))
			DebugPrint(".");

		// Set the TWI Master Mode Register
		twiPtr->TWI_MMR = ((TWSI_EEPROM_ADDRESS << 16) | AT91C_TWI_IADRSZ_2_BYTE ) & ~AT91C_TWI_MREAD;

		// Set TWI Internal Address Register
		twiPtr->TWI_IADR = ee_addr++;

		status = twiPtr->TWI_SR;

		twiPtr->TWI_THR = *(data_addr++);

		twiPtr->TWI_CR = AT91C_TWI_START;

		// Wait transfer is finished
		while (!(twiPtr->TWI_SR & AT91C_TWI_TXRDY));

		twiPtr->TWI_CR = AT91C_TWI_STOP;

		status = twiPtr->TWI_SR;

		// Wait transfer is finished
		while (!(twiPtr->TWI_SR & AT91C_TWI_TXCOMP));

		// wait for write operation to complete
		ReadEEPROM(ee_addr, &test_data, 1);
	}

	DebugPrint("\n\r");
}


/*
 * .KB_C_FN_DEFINITION_START
 * int eewrite_parse(int argc, char *argv[])
 *  This global function parses text from the command line.
 * .KB_C_FN_DEFINITION_END
 */
int eewrite_parse(int argc, char *argv[]) {

	unsigned eeAddr, srcAddr, size;

	if (argc < 4) {
		DebugPrint("Missing parameter");
		return (1);
	}

	p_ASCIIToHex(argv[1], &eeAddr);
	p_ASCIIToHex(argv[2], &srcAddr);
	p_ASCIIToHex(argv[3], &size);
	WriteEEPROM(eeAddr, (char*)srcAddr, size);

	return (0);
}


/*
 * .KB_C_FN_DEFINITION_START
 * void eewrite_help(void)
 *  This global displays help info for the corresponding command.
 * .KB_C_FN_DEFINITION_END
 */
void eewrite_help(void) {
	DebugPrint("\tprogram EEPROM using memory range\n\r");
	DebugPrint("\t\teewrite <ee_addr> <src_data_addr> <size>\n\r");
}


/*
 * .KB_C_FN_DEFINITION_START
 * void InitEEPROM(void)
 *  This global function initializes the EEPROM interface (TWI).  Intended
 * to be called a single time.
 * .KB_C_FN_DEFINITION_END
 */
void InitEEPROM(void) {

	AT91PS_TWI twiPtr = (AT91PS_TWI)AT91C_BASE_TWI;

	AT91PS_PIO pPio = (AT91PS_PIO)AT91C_BASE_PIOA;
	AT91PS_PMC pPMC = (AT91PS_PMC)AT91C_BASE_PMC;

	pPio->PIO_ASR =
		((unsigned)AT91C_PA25_TWD) | ((unsigned)AT91C_PA26_TWCK);
	pPio->PIO_BSR = 0;
	pPio->PIO_PDR =
		((unsigned)AT91C_PA25_TWD) | ((unsigned)AT91C_PA26_TWCK);

	pPio->PIO_MDDR = ~((unsigned)AT91C_PA25_TWD);
	pPio->PIO_MDER = (unsigned)AT91C_PA25_TWD;

	pPMC->PMC_PCER = ((unsigned) 1 << AT91C_ID_TWI);

	twiPtr->TWI_IDR = (unsigned) -1;
	twiPtr->TWI_CR = AT91C_TWI_SWRST;
	twiPtr->TWI_CR = AT91C_TWI_MSEN | AT91C_TWI_SVDIS;

	twiPtr->TWI_CWGR = SET_TWI_CLOCK;

	eeread_function.f_string = "eeread";
	eeread_function.parse_function = eeread_parse;
	eeread_function.help_function = eeread_help;
	RegisterFunction(&eeread_function);

	eewrite_function.f_string = "eewrite";
	eewrite_function.parse_function = eewrite_parse;
	eewrite_function.help_function = eewrite_help;
	RegisterFunction(&eewrite_function);
}


