/*******************************************************************************
 *
 * Filename: emac.c
 *
 * Instantiation of routines for MAC/ethernet functions.
 *
 * Revision information:
 *
 * 28AUG2004	kb_admin	initial creation
 * 08JAN2005	kb_admin	added tftp download
 *					also adapted from external sources
 *
 * BEGIN_KBDD_BLOCK
 * No warranty, expressed or implied, is included with this software.  It is
 * provided "AS IS" and no warranty of any kind including statutory or aspects
 * relating to merchantability or fitness for any purpose is provided.  All
 * intellectual property rights of others is maintained with the respective
 * owners.  This software is not copyrighted and is intended for reference
 * only.
 * END_BLOCK
 ******************************************************************************/

#include "AT91RM9200.h"
#include "emac.h"
#include "p_string.h"
#include "prompt.h"

/* ****************************** GLOBALS *************************************/

/* ********************** PRIVATE FUNCTIONS/DATA ******************************/

parse_function_t	localip_function, mac_function, mac2_function;
parse_function_t	serverip_function, tftp_function;

static unsigned localMACSet, serverMACSet, MAC_init;
static unsigned char localMACAddr[6], serverMACAddr[6];
static unsigned localIPSet, serverIPSet;
static unsigned char localIPAddr[4], serverIPAddr[4];
static unsigned short	serverPort, localPort;
static int	ackBlock;

static unsigned	lastAddress, lastSize;
static char *dlAddress;

static unsigned transmitBuffer[1024 / sizeof(unsigned)];
static unsigned tftpSendPacket[256 / sizeof(unsigned)];

receive_descriptor_t *p_rxBD;


/*
 * .KB_C_FN_DEFINITION_START
 * unsigned BuildIP(char *argv[])
 *  This private function packs the test IP info to an unsigned value.
 * .KB_C_FN_DEFINITION_END
 */
static unsigned BuildIP(char *argv[]) {
	unsigned	p_ip_addr, p_tmp_val;

	p_ASCIIToDec(argv[1], &p_ip_addr);
	p_ip_addr <<= 8;

	p_ASCIIToDec(argv[2], &p_tmp_val);
	p_ip_addr |= p_tmp_val;
	p_ip_addr <<= 8;

	p_ASCIIToDec(argv[3], &p_tmp_val);
	p_ip_addr |= p_tmp_val;
	p_ip_addr <<= 8;

	p_ASCIIToDec(argv[4], &p_tmp_val);
	p_ip_addr |= p_tmp_val;

	return (p_ip_addr);
}


/*
 * .KB_C_FN_DEFINITION_START
 * unsigned short IP_checksum(unsigned short *p, int len)
 *  This private function calculates the IP checksum for various headers.
 * .KB_C_FN_DEFINITION_END
 */
static unsigned short IP_checksum(unsigned short *p, int len) 
{
	unsigned	i, t;

	len &= ~1;

	for (i=0,t=0; i<len; i+=2, ++p)
		t += SWAP16(*p);

	t = (t & 0xffff) + (t >> 16);
	return (~t);									
}


/*
 * .KB_C_FN_DEFINITION_START
 * void GetServerAddress(void)
 *  This private function sends an ARP request to determine the server MAC.
 * .KB_C_FN_DEFINITION_END
 */
static void GetServerAddress(void) {

	arp_header_t	*p_ARP;

	p_ARP = (arp_header_t*)transmitBuffer;

	p_memset((char*)p_ARP->dest_mac, 0xFF, 6);

	p_memcpy((char*)p_ARP->src_mac, (char*)localMACAddr, 6);

	p_ARP->frame_type = SWAP16(PROTOCOL_ARP);
	p_ARP->hard_type  = SWAP16(1);
	p_ARP->prot_type  = SWAP16(PROTOCOL_IP);
	p_ARP->hard_size  = 6;
	p_ARP->prot_size  = 4;
	p_ARP->operation  = SWAP16(ARP_REQUEST);

	p_memcpy((char*)p_ARP->sender_mac, (char*)localMACAddr, 6);

	p_memcpy((char*)p_ARP->sender_ip, (char*)localIPAddr, 4);

	p_memset((char*)p_ARP->target_mac, 0, 6);

	p_memcpy((char*)p_ARP->target_ip, (char*)serverIPAddr, 4);

	// wait until transmit is available
	while (!(*AT91C_EMAC_TSR & AT91C_EMAC_BNQ)) ;

  	*AT91C_EMAC_TSR |= AT91C_EMAC_COMP;
	*AT91C_EMAC_TAR = (unsigned)transmitBuffer;
	*AT91C_EMAC_TCR = 0x40;
}


/*
 * .KB_C_FN_DEFINITION_START
 * void Send_TFTP_Packet(char *tftpData, unsigned tftpLength)
 *  This private function initializes and send a TFTP packet.
 * .KB_C_FN_DEFINITION_END
 */
static void Send_TFTP_Packet(char *tftpData, unsigned tftpLength)
{
	transmit_header_t	*macHdr = (transmit_header_t*)tftpSendPacket;
	ip_header_t		*ipHdr;
	udp_header_t		*udpHdr;
	unsigned		t_checksum;

	p_memcpy((char*)macHdr->dest_mac, (char*)serverMACAddr, 6);

	p_memcpy((char*)macHdr->src_mac, (char*)localMACAddr, 6);

	*(unsigned short*)&(macHdr->ip_tos) = SWAP16(PROTOCOL_IP);

	ipHdr = (ip_header_t*)&macHdr->packet_length;

	ipHdr->ip_v_hl = 0x45;
	ipHdr->ip_tos = 0;
	ipHdr->ip_len = SWAP16(28 + tftpLength);
	ipHdr->ip_id = 0;
	ipHdr->ip_off = SWAP16(0x4000);
	ipHdr->ip_ttl = 64;
	ipHdr->ip_p = PROTOCOL_UDP;
	ipHdr->ip_sum = 0;

	p_memcpy((char*)ipHdr->ip_src, (char*)localIPAddr, 4);

	p_memcpy((char*)ipHdr->ip_dst, (char*)serverIPAddr, 4);

	ipHdr->ip_sum = SWAP16(IP_checksum((unsigned short*)ipHdr, 20));

	udpHdr = (udp_header_t*)(ipHdr + 1);

	udpHdr->src_port  = SWAP16(localPort);
	udpHdr->dst_port  = SWAP16(serverPort);
	udpHdr->udp_len   = SWAP16(8 + tftpLength);
	udpHdr->udp_cksum = 0;

	p_memcpy((char*)udpHdr+8, tftpData, tftpLength);

	t_checksum = IP_checksum((unsigned short*)ipHdr + 6, (16 + tftpLength));

	t_checksum = (~t_checksum) & 0xFFFF;
	t_checksum += 25 + tftpLength;

	t_checksum = (t_checksum & 0xffff) + (t_checksum >> 16);
	t_checksum = (~t_checksum) & 0xFFFF;

	udpHdr->udp_cksum = SWAP16(t_checksum);

	while (!(*AT91C_EMAC_TSR & AT91C_EMAC_BNQ)) ;

  	*AT91C_EMAC_TSR |= AT91C_EMAC_COMP;
	*AT91C_EMAC_TAR = (unsigned)tftpSendPacket;
	*AT91C_EMAC_TCR = 42 + tftpLength;
}


/*
 * .KB_C_FN_DEFINITION_START
 * void TFTP_RequestFile(char *filename)
 *  This private function sends a RRQ packet to the server.
 * .KB_C_FN_DEFINITION_END
 */
static void TFTP_RequestFile(char *filename) {

	tftp_header_t	tftpHeader;
	char		*cPtr, *ePtr, *mPtr;
	unsigned	length;

	tftpHeader.opcode = SWAP16(TFTP_RRQ_OPCODE);

	cPtr = (char*)&(tftpHeader.block_num);

	ePtr = p_strcpy(cPtr, filename);
	mPtr = p_strcpy(ePtr, "octet");

	length = mPtr - cPtr;
	length += 2;
	
	Send_TFTP_Packet((char*)&tftpHeader, length);
}


/*
 * .KB_C_FN_DEFINITION_START
 * void TFTP_ACK_Data(char *data, unsigned short block_num, unsigned short len)
 *  This private function sends an ACK packet to the server.
 * .KB_C_FN_DEFINITION_END
 */
static void TFTP_ACK_Data(char *data, unsigned short block_num, unsigned short len) {

	tftp_header_t	tftpHeader;

	if (block_num == (ackBlock + 1)) {
		++ackBlock;
		p_memcpy(dlAddress, data, len);
		dlAddress += len;
		lastSize += len;
	}

	tftpHeader.opcode = SWAP16(TFTP_ACK_OPCODE);
	tftpHeader.block_num = SWAP16(block_num);
	Send_TFTP_Packet((char*)&tftpHeader, 4);

	if (len < 512)
		ackBlock = -2;
}


/*
 * .KB_C_FN_DEFINITION_START
 * void CheckForNewPacket(ip_header_t *pHeader)
 *  This private function polls for received ethernet packets and handles
 * any here.
 * .KB_C_FN_DEFINITION_END
 */
static void CheckForNewPacket(ip_header_t *pHeader) {

	unsigned short	*pFrameType, *pArpOp;
	unsigned	i;
	char		*pData;
	ip_header_t	*pIpHeader;
	arp_header_t	*p_ARP;
	int		process = 0;
	
	process = 0;	
	for (i = 0; i < MAX_RX_PACKETS; ++i) {				
		if(p_rxBD[i].address & 0x1) {
			process = 1;	
			(*AT91C_EMAC_RSR) |= (*AT91C_EMAC_RSR);
			break;				
		}
	}
		
	if (!process)
		return ;
						
	process = i;
		
	pFrameType = (unsigned short *) ((p_rxBD[i].address & 0xFFFFFFFC) + 12);
	pData      = (char *)(p_rxBD[i].address & 0xFFFFFFFC);

	switch (SWAP16(*pFrameType)) {

	case PROTOCOL_ARP:

		pArpOp = (unsigned short *) (pData + 20);
		p_ARP = (arp_header_t*)pData;

		if (SWAP16(*pArpOp) == ARP_REPLY) {

			// check if new server info is available
			if ((!serverMACSet) &&
				(!(p_memcmp((char*)p_ARP->sender_ip,
					(char*)serverIPAddr, 4)))) {

				serverMACSet = 1;

				p_memcpy((char*)serverMACAddr,
					(char*)p_ARP->sender_mac, 6);
			}
		}

		if (SWAP16(*pArpOp) == ARP_REQUEST) {

			// ARP REPLY operation
			*pArpOp =  SWAP16(ARP_REPLY);				

			// Fill the dest address and src address
			for (i = 0; i <6; i++) {
				// swap ethernet dest address and ethernet src address
				pData[i] = pData[i+6];
				pData[i+6] = localMACAddr[i];
				// swap sender ethernet address and target ethernet address
				pData[i+22] = localMACAddr[i];
				pData[i+32] = pData[i+6];
			}									

			// swap sender IP address and target IP address
			for (i = 0; i<4; i++) {				
				pData[i+38] = pData[i+28];
				pData[i+28] = localIPAddr[i];
			}

			if (!(*AT91C_EMAC_TSR & AT91C_EMAC_BNQ)) break;

		  	*AT91C_EMAC_TSR |= AT91C_EMAC_COMP;
			*AT91C_EMAC_TAR = (unsigned)pData;
 			*AT91C_EMAC_TCR = 0x40;
		}
	break;
		
	case PROTOCOL_IP:
		pIpHeader = (ip_header_t*)(pData + 14);			
		p_memcpy((char*)pHeader, (char*)pIpHeader,sizeof(ip_header_t));
		
		switch(pIpHeader->ip_p) {

		case PROTOCOL_UDP:
		{
			udp_header_t	*udpHdr;
			tftp_header_t	*tftpHdr;

			udpHdr = (udp_header_t*)((char*)pIpHeader+20);
			tftpHdr = (tftp_header_t*)((char*)udpHdr + 8);

			if (SWAP16(udpHdr->dst_port) != localPort)
				break;

			if (SWAP16(tftpHdr->opcode) != TFTP_DATA_OPCODE)
				break;

			if (ackBlock == -1) {
				if (SWAP16(tftpHdr->block_num) != 1)
						break;
					serverPort = SWAP16(udpHdr->src_port);
					ackBlock = 0;
				}

			if (serverPort != SWAP16(udpHdr->src_port))
				break;

			TFTP_ACK_Data(tftpHdr->data,
				SWAP16(tftpHdr->block_num),
				SWAP16(udpHdr->udp_len) - 12);
		}
		break;

		default:
			break;
		}
	break;
						
	default:
		break;
	}
	p_rxBD[process].address &= ~0x01;
}


/*
 * .KB_C_FN_DEFINITION_START
 * unsigned short AT91F_MII_ReadPhy (AT91PS_EMAC pEmac, unsigned char addr)
 *  This private function reads the PHY device.
 * .KB_C_FN_DEFINITION_END
 */
static unsigned short AT91F_MII_ReadPhy (AT91PS_EMAC pEmac, unsigned char addr) {

	unsigned value = 0x60020000 | (addr << 18);

	pEmac->EMAC_CTL |= AT91C_EMAC_MPE;
	pEmac->EMAC_MAN = value;
  	while(!((pEmac->EMAC_SR) & AT91C_EMAC_IDLE));
	pEmac->EMAC_CTL &= ~AT91C_EMAC_MPE;
	return (pEmac->EMAC_MAN & 0x0000ffff);
}


/*
 * .KB_C_FN_DEFINITION_START
 * void MII_GetLinkSpeed(AT91PS_EMAC pEmac)
 *  This private function determines the link speed set by the PHY.
 * .KB_C_FN_DEFINITION_END
 */
static void MII_GetLinkSpeed(AT91PS_EMAC pEmac) {

	unsigned short stat2; 
	unsigned update = 0;
  
	stat2 = AT91F_MII_ReadPhy(pEmac, MII_STS2_REG);

	if (!(stat2 & 0x400)) {
		return ;

	} else if (stat2 & 0x4000) {

		update |= AT91C_EMAC_SPD;

		if (stat2 & 0x200) {
			update |= AT91C_EMAC_FD;
		}

	} else if (stat2 & 0x200) {
		update |= AT91C_EMAC_FD;
	}

	pEmac->EMAC_CFG =
		(pEmac->EMAC_CFG & ~(AT91C_EMAC_SPD | AT91C_EMAC_FD)) | update;
}


/*
 * .KB_C_FN_DEFINITION_START
 * void AT91F_EmacEntry(void)
 *  This private function initializes the EMAC on the chip.
 * .KB_C_FN_DEFINITION_END
 */
void AT91F_EmacEntry(void) {

	unsigned	i;
	char		*pRxPacket = (char*)RX_DATA_START;
	AT91PS_EMAC	pEmac = AT91C_BASE_EMAC;

	for (i = 0; i < MAX_RX_PACKETS; ++i) {

		p_rxBD[i].address = (unsigned)pRxPacket;
		p_rxBD[i].size = 0;
		pRxPacket += RX_PACKET_SIZE;
	}
	
	// Set the WRAP bit at the end of the list descriptor
	p_rxBD[MAX_RX_PACKETS-1].address |= 0x02;

	pEmac->EMAC_CTL  = 0;

	if(!(pEmac->EMAC_SR & AT91C_EMAC_LINK))
		MII_GetLinkSpeed(pEmac);

	// the sequence write EMAC_SA1L and write EMAC_SA1H must be respected
	pEmac->EMAC_SA1L = ((unsigned)localMACAddr[2] << 24) | ((unsigned)localMACAddr[3] << 16) | ((int)localMACAddr[4] << 8) | localMACAddr[5];
	pEmac->EMAC_SA1H = ((unsigned)localMACAddr[0] << 8) | localMACAddr[1];

	pEmac->EMAC_RBQP = (unsigned) p_rxBD;
	pEmac->EMAC_RSR  |= (AT91C_EMAC_OVR | AT91C_EMAC_REC | AT91C_EMAC_BNA);
	pEmac->EMAC_CFG  |= AT91C_EMAC_CAF;
	pEmac->EMAC_CFG  = (pEmac->EMAC_CFG & ~(AT91C_EMAC_CLK)) |
		AT91C_EMAC_CLK_HCLK_32;
	pEmac->EMAC_CTL  |= (AT91C_EMAC_TE | AT91C_EMAC_RE);

	pEmac->EMAC_TAR = (unsigned)transmitBuffer;
}	


/* ************************** GLOBAL FUNCTIONS ********************************/


/*
 * .KB_C_FN_DEFINITION_START
 * void SetMACAddress(unsigned low_address, unsigned high_address)
 *  This global function sets the MAC address.  low_address is the first
 * four bytes while high_address is the last 2 bytes of the 48-bit value.
 * .KB_C_FN_DEFINITION_END
 */
void SetMACAddress(unsigned low_address, unsigned high_address) {

	AT91PS_EMAC	pEmac = AT91C_BASE_EMAC;
	AT91PS_PMC	pPMC = AT91C_BASE_PMC;

	/* enable the peripheral clock before using EMAC */
	pPMC->PMC_PCER = ((unsigned) 1 << AT91C_ID_EMAC);

	pEmac->EMAC_SA1L = low_address;
	pEmac->EMAC_SA1H = (high_address & 0x0000ffff);

	localMACAddr[0] = (low_address >>  0) & 0xFF;
	localMACAddr[1] = (low_address >>  8) & 0xFF;
	localMACAddr[2] = (low_address >> 16) & 0xFF;
	localMACAddr[3] = (low_address >> 24) & 0xFF;
	localMACAddr[4] = (high_address >> 0) & 0xFF;
	localMACAddr[5] = (high_address >> 8) & 0xFF;

	localMACSet = 1;

	// low_address  & 0x000000ff = first byte in address
	// low_address  & 0x0000ff00 = next
	// low_address  & 0x00ff0000 = next
	// low_address  & 0xff000000 = next
	// high_address & 0x000000ff = next
	// high_address & 0x0000ff00 = last byte in address
}


/*
 * .KB_C_FN_DEFINITION_START
 * void SetServerIPAddress(unsigned address)
 *  This global function sets the IP of the TFTP download server.
 * .KB_C_FN_DEFINITION_END
 */
void SetServerIPAddress(unsigned address) {

	// force update in case the IP has changed
	serverMACSet = 0;

	serverIPAddr[0] = (address >> 24) & 0xFF;
	serverIPAddr[1] = (address >> 16) & 0xFF;
	serverIPAddr[2] = (address >>  8) & 0xFF;
	serverIPAddr[3] = (address >>  0) & 0xFF;

	serverIPSet = 1;
}


/*
 * .KB_C_FN_DEFINITION_START
 * void SetLocalIPAddress(unsigned address)
 *  This global function sets the IP of this module.
 * .KB_C_FN_DEFINITION_END
 */
void SetLocalIPAddress(unsigned address) {

	// force update in case the IP has changed
	serverMACSet = 0;

	localIPAddr[0] = (address >> 24) & 0xFF;
	localIPAddr[1] = (address >> 16) & 0xFF;
	localIPAddr[2] = (address >>  8) & 0xFF;
	localIPAddr[3] = (address >>  0) & 0xFF;

	localIPSet = 1;
}


/*
 * .KB_C_FN_DEFINITION_START
 * void TFTP_Download(unsigned address, char *filename)
 *  This global function initiates and processes a tftp download request.
 * The server IP, local IP, local MAC must be set before this function is
 * executed.
 * .KB_C_FN_DEFINITION_END
 */
void TFTP_Download(unsigned address, char *filename)
{
	ip_header_t 	IpHeader;
	AT91PS_EMAC 	pEmac = (AT91PS_EMAC)AT91C_BASE_EMAC;
	unsigned	thisSeconds, running, state;
	int		timeout, tickUpdate;

	if (!address) {
		// report last transfer information
		DebugPrint("\n\r -- Last tftp transfer info --\n\r");
		DebugPrint("     address: ");
		DebugPrintHex(8, lastAddress);
		DebugPrint("\n\r     size: ");
		DebugPrintHex(8, lastSize);
		DebugPrint("\n\r");
		return ;
	}

	if ((!localMACSet) || (!localIPSet) || (!serverIPSet))
		return ;

	if (!MAC_init) {
		AT91C_BASE_PMC->PMC_PCER = ((unsigned) 1 << AT91C_ID_EMAC);

		AT91C_BASE_PIOA->PIO_ASR = 
			((unsigned) AT91C_PA14_ERXER   ) |
			((unsigned) AT91C_PA12_ERX0    ) |
			((unsigned) AT91C_PA13_ERX1    ) |
			((unsigned) AT91C_PA8_ETXEN   ) |
			((unsigned) AT91C_PA16_EMDIO   ) |
			((unsigned) AT91C_PA9_ETX0    ) |
			((unsigned) AT91C_PA10_ETX1    ) |
			((unsigned) AT91C_PA11_ECRS_ECRSDV) |
			((unsigned) AT91C_PA15_EMDC    ) |
			((unsigned) AT91C_PA7_ETXCK_EREFCK);
		AT91C_BASE_PIOA->PIO_BSR = 0;
		AT91C_BASE_PIOA->PIO_PDR = 
			((unsigned) AT91C_PA14_ERXER   ) |
			((unsigned) AT91C_PA12_ERX0    ) |
			((unsigned) AT91C_PA13_ERX1    ) |
			((unsigned) AT91C_PA8_ETXEN   ) |
			((unsigned) AT91C_PA16_EMDIO   ) |
			((unsigned) AT91C_PA9_ETX0    ) |
			((unsigned) AT91C_PA10_ETX1    ) |
			((unsigned) AT91C_PA11_ECRS_ECRSDV) |
			((unsigned) AT91C_PA15_EMDC    ) |
			((unsigned) AT91C_PA7_ETXCK_EREFCK);

		AT91C_BASE_PIOB->PIO_ASR = 0;
		AT91C_BASE_PIOB->PIO_BSR =
			((unsigned) AT91C_PB12_ETX2) |
			((unsigned) AT91C_PB13_ETX3) |
			((unsigned) AT91C_PB14_ETXER) |
			((unsigned) AT91C_PB15_ERX2) |
			((unsigned) AT91C_PB16_ERX3) |
			((unsigned) AT91C_PB17_ERXDV) |
			((unsigned) AT91C_PB18_ECOL) |
			((unsigned) AT91C_PB19_ERXCK);
		AT91C_BASE_PIOB->PIO_PDR =
			((unsigned) AT91C_PB12_ETX2) |
			((unsigned) AT91C_PB13_ETX3) |
			((unsigned) AT91C_PB14_ETXER) |
			((unsigned) AT91C_PB15_ERX2) |
			((unsigned) AT91C_PB16_ERX3) |
			((unsigned) AT91C_PB17_ERXDV) |
			((unsigned) AT91C_PB18_ECOL) |
			((unsigned) AT91C_PB19_ERXCK);

		MAC_init = 1;
	}

	AT91F_EmacEntry();

	GetServerAddress();

	lastAddress = address;
	dlAddress = (char*)address;
	lastSize = 0;
	running = 1;
	state = TFTP_WAITING_SERVER_MAC;
	timeout = 20;
	thisSeconds = GetSeconds();
	serverPort = 69;
	++localPort;
	ackBlock = -1;

	while (running && timeout) {

		CheckForNewPacket(&IpHeader);

		tickUpdate = 0;

		if (thisSeconds != GetSeconds()) {
			tickUpdate = 1;
			--timeout;
			thisSeconds = GetSeconds();
		}

		switch (state) {

			case TFTP_WAITING_SERVER_MAC:
				if (serverMACSet) {
					state = TFTP_SEND_REQUEST;
					break;
				}

				if (tickUpdate)
					GetServerAddress();
				break;

			case TFTP_SEND_REQUEST:
				// send request for file
				if (ackBlock != -1) {
					state = TFTP_GET_DATA;
					break;
				}

				if (tickUpdate)
					TFTP_RequestFile(filename);
				break;

			case TFTP_GET_DATA:
				// receiving data
				if (ackBlock == -2) {
					state = TFTP_COMPLETE;
					break;
				}
				break;

			case TFTP_COMPLETE:
			default:
				running = 0;
				break;
		}
	}
}


/*
 * .KB_C_FN_DEFINITION_START
 * int localip_function_parse(int argc, char *argv[])
 *  This global function parses text from the command line.
 * .KB_C_FN_DEFINITION_END
 */
int localip_function_parse(int argc, char *argv[]) {

	if (argc > 4)
		SetLocalIPAddress(BuildIP(argv));
	return (0);
}


/*
 * .KB_C_FN_DEFINITION_START
 * void localip_function_help(void)
 *  This global displays help info for the corresponding command.
 * .KB_C_FN_DEFINITION_END
 */
void localip_function_help(void) {
	DebugPrint("\tset localip\n\r");
	DebugPrint("\t\tip a b c d (decimal)\n\r");
}


/*
 * .KB_C_FN_DEFINITION_START
 * int mac_function_parse(int argc, char *argv[])
 *  This global function parses text from the command line.
 * .KB_C_FN_DEFINITION_END
 */
int mac_function_parse(int argc, char *argv[]) {

	unsigned	u_value, low_addr, high_addr;

	if (argc > 6) {

		p_ASCIIToHex(argv[4], &low_addr);
		low_addr <<= 8;

		p_ASCIIToHex(argv[3], &u_value);
		low_addr |= u_value;
		low_addr <<= 8;

		p_ASCIIToHex(argv[2], &u_value);
		low_addr |= u_value;
		low_addr <<= 8;

		p_ASCIIToHex(argv[1], &u_value);
		low_addr |= u_value;

		p_ASCIIToHex(argv[6], &high_addr);
		high_addr <<= 8;

		p_ASCIIToHex(argv[5], &u_value);
		high_addr |= u_value;

		SetMACAddress(low_addr, high_addr);
	}
	return (0);
}


/*
 * .KB_C_FN_DEFINITION_START
 * void mac_function_help(void)
 *  This global displays help info for the corresponding command.
 * .KB_C_FN_DEFINITION_END
 */
void mac_function_help(void) {
	DebugPrint("\tset local mac address\n\r");
	DebugPrint("\t\t<m | mac> a b c d e f (hex)\n\r");
}


/*
 * .KB_C_FN_DEFINITION_START
 * int serverip_function_parse(int argc, char *argv[])
 *  This global function parses text from the command line.
 * .KB_C_FN_DEFINITION_END
 */
int serverip_function_parse(int argc, char *argv[]) {

	if (argc > 4)
		SetServerIPAddress(BuildIP(argv));
	return (0);
}


/*
 * .KB_C_FN_DEFINITION_START
 * void serverip_function_help(void)
 *  This global displays help info for the corresponding command.
 * .KB_C_FN_DEFINITION_END
 */
void serverip_function_help(void) {
	DebugPrint("\tset server ip\n\r");
	DebugPrint("\t\tserver_ip a b c d (decimal)\n\r");
}


/*
 * .KB_C_FN_DEFINITION_START
 * int tftp_function_parse(int argc, char *argv[])
 *  This global function parses text from the command line.
 * .KB_C_FN_DEFINITION_END
 */
int tftp_function_parse(int argc, char *argv[]) {

	unsigned address = 0;

	if (argc > 2)
		p_ASCIIToHex(argv[1], &address);

	DebugPrint("Starting tftp download of file: ");
	DebugPrint(argv[2]);
	DebugPrint(" at: ");
	DebugPrintHex(8, address);
	DebugPrint("\n\r");
	TFTP_Download(address, argv[2]);

	return (0);
}


/*
 * .KB_C_FN_DEFINITION_START
 * void tftp_function_help(void)
 *  This global displays help info for the corresponding command.
 * .KB_C_FN_DEFINITION_END
 */
void tftp_function_help(void) {
	DebugPrint("\ttftp download\n\r");
	DebugPrint("\t\ttfpt dest_addr filename\n\r");
}


/*
 * .KB_C_FN_DEFINITION_START
 * void EMAC_Init(void)
 *  This global function initializes variables used in tftp transfers.
 * .KB_C_FN_DEFINITION_END
 */
void EMAC_Init(void) {

	p_rxBD = (receive_descriptor_t*)RX_BUFFER_START;
	localMACSet = 0;
	serverMACSet = 0;
	localIPSet = 0;
	serverIPSet = 0;
	localPort = 0x8002;
	lastAddress = 0;
	lastSize = 0;
	MAC_init = 0;

	localip_function.f_string = "ip";
	localip_function.parse_function = localip_function_parse;
	localip_function.help_function = localip_function_help;
	RegisterFunction(&localip_function);

	mac_function.f_string = "mac";
	mac_function.parse_function = mac_function_parse;
	mac_function.help_function = mac_function_help;
	RegisterFunction(&mac_function);

	mac2_function.f_string = "m";
	mac2_function.parse_function = mac_function_parse;
	mac2_function.help_function = 0;
	RegisterFunction(&mac2_function);

	serverip_function.f_string = "server_ip";
	serverip_function.parse_function = serverip_function_parse;
	serverip_function.help_function = serverip_function_help;
	RegisterFunction(&serverip_function);

	tftp_function.f_string = "tftp";
	tftp_function.parse_function = tftp_function_parse;
	tftp_function.help_function = tftp_function_help;
	RegisterFunction(&tftp_function);
}
