/*******************************************************************************
 *
 * Filename: LCD.c
 *
 * Instantiation of LCD support routines
 *
 * Revision information:
 *
 * 15FEB2005	kb_admin	initial creation
 *
 * BEGIN_KBDD_BLOCK
 * No warranty, expressed or implied, is included with this software.  It is
 * provided "AS IS" and no warranty of any kind including statutory or aspects
 * relating to merchantability or fitness for any purpose is provided.  All
 * intellectual property rights of others is maintained with the respective
 * owners.  This software is not copyrighted and is intended for reference
 * only.
 * END_BLOCK
 ******************************************************************************/

#include "lcd.h"
#include "AT91RM9200.h"
#include "prompt.h"
#include "array_bytes.c"

/* ****************************** GLOBALS *************************************/


/* ********************** PRIVATE FUNCTIONS/DATA ******************************/


parse_function_t	lcd_function;

unsigned	invert;


#define	WRITE_COMMAND(x)	((*(char*)0x30000000) = x)
#define	READ_COMMAND		(*(char*)0x30000000)
#define	WRITE_DATA(x)		((*(char*)0x30010000) = x)
#define	READ_DATA		(*(char*)0x30010000)


#define	RESET_DISPLAY		0xE2
#define	LCD_BIAS_1_9		0xA2
#define	ADC_SELECT_REVERSE	0xA1
#define	COMMON_OUTPUT_NORMAL	0xC0
#define	V5_RESISTOR_RATIO	0x26
#define	ELECTRONIC_VOLUME_SET	0x81
#define	ELECTRONIC_VOLUME_INIT	0x20
#define	POWER_CONTROL_SET	0x28
#define	VOLTAGE_REGULATOR	0x02
#define	VOLTAGE_FOLLOWER	0x01
#define	BOOSTER_CIRCUIT		0x04
#define	DISPLAY_ON		0xAF
#define	START_LINE_SET		0x40
#define	PAGE_ADDRESS_SET	0xB0
#define	COLUMN_ADDRESS_HIGH	0x10
#define	COLUMN_ADDRESS_LOW	0x00
#define	RESISTOR_RATIO_START	0x20

#define	NUM_OF_PAGES		8
#define	NUM_OF_COLUMNS		128


#define	LCD_BACKLIGHT_PIN	AT91C_PIO_PC23

/*
 * .KB_C_FN_DEFINITION_START
 * void WritePattern(unsigned width)
 *  This private functions write a checkboard pattern to the display.
 * .KB_C_FN_DEFINITION_END
 */
void WritePattern(unsigned width) {

	unsigned	i, j, page, pattern;

	switch (width) {
		case 1:
			pattern = 0x55;
			break;
		case 2:
			pattern = 0x33;
			break;
		case 4:
			pattern = 0x0F;
			break;
		default:
			width = 8;
			pattern = 0x00;
	}

	for (page = 0; page < NUM_OF_PAGES; ++page) {
		WRITE_COMMAND(START_LINE_SET);
		WRITE_COMMAND(PAGE_ADDRESS_SET + page);
		WRITE_COMMAND(COLUMN_ADDRESS_HIGH);
		WRITE_COMMAND(COLUMN_ADDRESS_LOW);

		if (width == 8)
			pattern = ~pattern & 0xFF;

		for (i = 0; i < NUM_OF_COLUMNS/width; ++i) {

			for (j = 0; j < width; ++j)
				WRITE_DATA(pattern);

			pattern = ~pattern & 0xFF;
		}
	}

	WRITE_COMMAND(DISPLAY_ON);
}


/*
 * .KB_C_FN_DEFINITION_START
 * void VaryContrast(void)
 *  This private functions varies the contrast to find optimal values.
 * .KB_C_FN_DEFINITION_END
 */
/*
void VaryContrast(void) {

	char	i, j;

	for (i = 0; i < 8; ++i) {

		WRITE_COMMAND(RESISTOR_RATIO_START + i);

		for (j = 0; j < 64; ++j) {

			WRITE_COMMAND(ELECTRONIC_VOLUME_SET);
			WRITE_COMMAND(j);
			DebugPrint("\n\rR, E: ");
			DebugPrintHex(2, i);
			DebugPrint(",");
			DebugPrintHex(2, j);
			while (!DebugGetchar(0)) ;
		}
	}
	DebugPrint("\n\r");
}
*/


/*
 * .KB_C_FN_DEFINITION_START
 * void WriteLogo(void)
 *  This private functions writes the KB logo to the display.
 * .KB_C_FN_DEFINITION_END
 */
void WriteLogo(void) {

	unsigned	pattern, page, i;
	char		*cPtr;

	cPtr = (char*)KBlogo;

	for (page = 0; page < NUM_OF_PAGES; ++page) {
		WRITE_COMMAND(START_LINE_SET);
		WRITE_COMMAND(PAGE_ADDRESS_SET + page);
		WRITE_COMMAND(COLUMN_ADDRESS_HIGH);
		WRITE_COMMAND(COLUMN_ADDRESS_LOW);

		for (i = 0; i < NUM_OF_COLUMNS; ++i) {
			pattern = *cPtr++;

			if (invert == 0x10)
				pattern ^= 0xFF;

			WRITE_DATA(pattern);
		}
	}

	WRITE_COMMAND(DISPLAY_ON);


//	VaryContrast();
}


/*
 * .KB_C_FN_DEFINITION_START
 * void SetLCDBacklight(int enable)
 *  This private functions enables the LCD backlight if enable is
 *  non-zero.  Otherwise, the LCD backlight is disabled.
 * .KB_C_FN_DEFINITION_END
 */
void SetLCDBacklight(int enable) {

	if (enable) {

		AT91C_BASE_PIOC->PIO_SODR = LCD_BACKLIGHT_PIN;

	} else {

		AT91C_BASE_PIOC->PIO_CODR = LCD_BACKLIGHT_PIN;
	}
}


/* ************************** GLOBAL FUNCTIONS ********************************/


/*
 * .KB_C_FN_DEFINITION_START
 * int lcd_parse(int argc, char *argv[])
 *  This global function parses text from the command line.
 * .KB_C_FN_DEFINITION_END
 */
int lcd_parse(int argc, char *argv[]) {

	if (argc < 2) {

		if (invert < 0x10) {

			WritePattern(invert);

		} else {

			WriteLogo();
		}

		invert <<= 1;

		if (invert > 0x20)
			invert = 1;

	} else if (!p_strcmp(argv[1], "on")) {

		SetLCDBacklight(1);

	} else {

		SetLCDBacklight(0);

	}

	return (0);
}


/*
 * .KB_C_FN_DEFINITION_START
 * void lcd_help(void)
 *  This global displays help info for the corresponding command.
 * .KB_C_FN_DEFINITION_END
 */
void lcd_help(void) {
	DebugPrint("\tGenerate LCD display pattern\n\r");
	DebugPrint("\t\tlcd\n\r");
	DebugPrint("\tEnable (disable) LCD backlight\n\r");
	DebugPrint("\t\tlcd <on|off>\n\r");
}


/*
 * .KB_C_FN_DEFINITION_START
 * void InitLCD(void)
 *  This global function initializes the LCD interface.  The timing
 *  parameters allow for ample margin.  To increase performance, tune
 *  these values.
 * .KB_C_FN_DEFINITION_END
 */
void InitLCD(void) {

	unsigned	value;

	invert = 1;

	// setup LCD access (allow ample margin)
	// 18 wait states, 1 setup, 1 hold, 1 float for 8-bit device
	AT91C_BASE_SMC2->SMC2_CSR[2] =
		AT91C_SMC2_WSEN |
		(18 & AT91C_SMC2_NWS) |
		((1 << 8) & AT91C_SMC2_TDF) |
		AT91C_SMC2_DBW_8 |
		((1 << 24) & AT91C_SMC2_RWSETUP) |
		((1 << 29) & AT91C_SMC2_RWHOLD);



	lcd_function.f_string = "lcd";
	lcd_function.parse_function = lcd_parse;
	lcd_function.help_function = lcd_help;
	RegisterFunction(&lcd_function);

	// release the reset line by reading the device
	value = READ_COMMAND;
	value = READ_COMMAND;

	// newer versions use a PIO line for the reset signal
	AT91C_BASE_PIOC->PIO_PER = AT91C_PIO_PC22;
	AT91C_BASE_PIOC->PIO_OER = AT91C_PIO_PC22;
	AT91C_BASE_PIOC->PIO_SODR = AT91C_PIO_PC22;

	// initialize the device within 5ms
	WRITE_COMMAND(RESET_DISPLAY);
	WRITE_COMMAND(LCD_BIAS_1_9);
	WRITE_COMMAND(ADC_SELECT_REVERSE);
	WRITE_COMMAND(COMMON_OUTPUT_NORMAL);
	WRITE_COMMAND(V5_RESISTOR_RATIO);
	WRITE_COMMAND(ELECTRONIC_VOLUME_SET);
	WRITE_COMMAND(ELECTRONIC_VOLUME_INIT);
	WRITE_COMMAND(POWER_CONTROL_SET | VOLTAGE_REGULATOR | VOLTAGE_FOLLOWER | BOOSTER_CIRCUIT);
	WRITE_COMMAND(DISPLAY_ON);

	// in test, either (0x04, 0x33 ) or (0x05, 0x27)
	WRITE_COMMAND(RESISTOR_RATIO_START + 4);
	WRITE_COMMAND(ELECTRONIC_VOLUME_SET);
	WRITE_COMMAND(0x33);

	// disable the LCD backlight by default
	AT91C_BASE_PIOC->PIO_PER = LCD_BACKLIGHT_PIN;
	AT91C_BASE_PIOC->PIO_OER = LCD_BACKLIGHT_PIN;

	WriteLogo();
	SetLCDBacklight(0);
}
