/*-
 * Copyright (c) 2006 M. Warner Losh <imp@village.org> 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Core driver for the SD Standard Host Controller interface.  This
 * interface has been deduced from the Linux mmc driver, the OpenBSD
 * sdhc driver and the Simplified SD standard.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/bus.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/module.h>
#include <sys/systm.h>

#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/rman.h> 

#include <dev/pci/pcivar.h>
#include <dev/pci/pcireg.h>

#include <dev/sdhc/sdhcvar.h>
#include <dev/sdhc/sdhcreg.h>

devclass_t sdhc_devclass;

int
sdhc_read_ivar(device_t dev, device_t child, int which, uintptr_t *result)
{
	struct sdhc_ivars *ivars = device_get_softc(dev);

	switch (which) {
	case SDHC_IVAR_RESOURCE:
		*(struct resource **)result = ivars->res;
		return (0);
	}

	return (ENOENT);
}

int
sdhc_write_ivar(device_t dev, device_t child, int which, uintptr_t value)
{
	struct sdhc_ivars *ivars = device_get_softc(dev);

	switch (which) {
	case SDHC_IVAR_RESOURCE:
		ivars->res = *(struct resource **)value;
		return (0);
	}

	return (ENOENT);
}

struct resource *
sdhc_alloc_resource(device_t dev, device_t child, int type, int *rid,
    u_long start, u_long end, u_long count, u_int flags)
{
	struct resource *r;

	if (*rid != 0 || type != SYS_RES_MEMORY ||
	    device_get_parent(child) != dev)
		return (NULL);
	r = sdhc_get_resource(child);
	rman_set_device(r, child);

	return (r);
}
		
int
sdhc_release_resource(device_t dev, device_t child, int type, int rid,
    struct resource *r)
{
	if (rid != 0 || type != SYS_RES_MEMORY ||
	    device_get_parent(child) != dev || rman_get_device(r) != dev)
		return (EINVAL);
	rman_set_device(r, dev);

	return (0);
}

int
sdhc_attach(device_t dev)
{
	return (0);
}

int
sdhc_detach(device_t dev)
{
	struct sdhc_softc *sc = device_get_softc(dev);
	int numdevs, tmp, i;
	device_t *devlist;

	bus_generic_detach(dev);

	for (i = 0; i < sc->nres; i++)
		bus_free_resource(dev, SYS_RES_MEMORY, sc->res[i]);

	// Since we create these devices in attach, we have to delete them
	// in detach.
	device_get_children(dev, &devlist, &numdevs);
	for (tmp = 0; tmp < numdevs; tmp++)
		device_delete_child(dev, devlist[tmp]);
	free(devlist, M_TEMP);

	if (sc->intrhand)
		bus_teardown_intr(dev, sc->irq_res, sc->intrhand);
	if (sc->irq_res)
		bus_free_resource(dev, SYS_RES_IRQ, sc->irq_res);
	return (0);
}

struct sdhc_host_softc 
{
	int dummy;
};

static int
sdhc_host_probe(device_t dev)
{
	return (0);
}

static int
sdhc_host_attach(device_t dev)
{
	device_add_child(dev, "mmcbus", -1);
	return (0);
}

static int
sdhc_host_detach(device_t dev)
{
	int numdevs, tmp;
	device_t *devlist;

	// Since we create these devices in attach, we have to delete them
	// in detach.
	device_get_children(dev, &devlist, &numdevs);
	for (tmp = 0; tmp < numdevs; tmp++)
		device_delete_child(dev, devlist[tmp]);
	free(devlist, M_TEMP);

	return (0);
	
}

static device_method_t sdhc_host_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,			sdhc_host_probe),
	DEVMETHOD(device_attach,		sdhc_host_attach),
	DEVMETHOD(device_detach,		sdhc_host_detach),

	{ 0, 0 }
};

static driver_t sdhc_host_driver = {
	"sdhc_host",
	sdhc_host_methods,
	sizeof(struct sdhc_host_softc),
};

devclass_t sdhc_host_devclass;

DRIVER_MODULE(sdhc_host, sdhc, sdhc_host_driver, sdhc_host_devclass, 0, 0);
