/*-
 * Copyright (c) 2006 M. Warner Losh <imp@village.org> 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * PCI and Cardbus attachments for the SD Standard Host Controller
 * interface.  This interface has been deduced from the Linux mmc
 * driver, the OpenBSD sdhc driver and the Simplified SD standard.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/bus.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/module.h>
#include <sys/systm.h>

#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/rman.h> 

#include <dev/pci/pcivar.h>
#include <dev/pci/pcireg.h>

#include <dev/sdhc/sdhcvar.h>
#include <dev/sdhc/sdhcreg.h>

static void sdhc_pci_intr(void *arg);

static int
sdhc_pci_probe(device_t dev)
{
	if (pci_get_class(dev) != PCIC_BASEPERIPH ||
	    pci_get_subclass(dev) != PCIS_BASEPERIPH_SDHC)
		return (ENXIO);
	device_set_desc(dev, "SD Standard Host Controller");
	return (BUS_PROBE_SPECIFIC);
}

static int
sdhc_pci_attach(device_t dev)
{
	struct sdhc_softc *sc = device_get_softc(dev);
	device_t child;
	struct sdhc_ivars *ivars;
	struct resource *res;
	int i, rid, err;

	if (pci_get_progif(dev) == SDHC_PCI_INTERFACE_DMA)
		sc->flags |= SDHC_FLAG_HASDMA;
	/* Map and establish the interrupt. */
	rid = 0;
	sc->irq_res = bus_alloc_resource_any(dev, SYS_RES_IRQ, &rid,
	    RF_SHAREABLE | RF_ACTIVE);
	if (sc->irq_res == NULL) {
		device_printf(dev, "Unable to map IRQ...\n");
		goto errout;
	}

	for (i = SDHC_RID_MIN; i <= SDHC_RID_MAX; i += sizeof(uint32_t)) {
		res = bus_alloc_resource_any(dev, SYS_RES_MEMORY, &i, RF_ACTIVE);
		if (res == NULL)
			continue;
		sc->res[sc->nres++] = res;
		child = device_add_child(dev, "sdhc_host", -1);
		if (child == NULL) {
			device_printf(child, "Can't add host slot rid %#x\n",
			    i);
			bus_release_resource(dev, SYS_RES_MEMORY, i, res);
			continue;
		}
		ivars = (struct sdhc_ivars *) malloc(sizeof *ivars, M_DEVBUF,
		    M_WAITOK);
		ivars->res = res;
		device_set_ivars(child, ivars);
	}

	err = sdhc_attach(dev);
	if (err)
		goto errout;

	if (bus_setup_intr(dev, sc->irq_res, INTR_TYPE_AV | INTR_MPSAFE,
	    sdhc_pci_intr, sc, &sc->intrhand)) {
		device_printf(dev, "couldn't establish interrupt");
		goto errout;
	}
	return (0);
errout:;
	sdhc_detach(dev);
	return (ENXIO);
}

static void
sdhc_pci_intr(void *arg)
{
}

static device_method_t sdhc_pci_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,			sdhc_pci_probe),
	DEVMETHOD(device_attach,		sdhc_pci_attach),
	DEVMETHOD(device_detach,		sdhc_detach),

	/* bus methods */
	DEVMETHOD(bus_read_ivar,		sdhc_read_ivar),
	DEVMETHOD(bus_write_ivar,		sdhc_write_ivar),
	DEVMETHOD(bus_alloc_resource,		sdhc_alloc_resource),
	DEVMETHOD(bus_release_resource,		sdhc_release_resource),

	{ 0, 0 }
};

static driver_t sdhc_pci_driver = {
	"sdhc",
	sdhc_pci_methods,
	sizeof(struct sdhc_softc),
};

DRIVER_MODULE(sdhc, pci, sdhc_pci_driver, sdhc_devclass, 0, 0);
DRIVER_MODULE(sdhc, cardbus, sdhc_pci_driver, sdhc_devclass, 0, 0);
MODULE_DEPEND(sdhc, pci, 1, 1, 1);
MODULE_DEPEND(sdhc, cardbus, 1, 1, 1);
