/*-
 * Copyright 2005 John-Mark Gurney
 * Copyright (c) 2003 Marcel Moolenaar
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD: src/sys/dev/uart/uart_dev_ns8250.c,v 1.14 2005/01/06 01:43:26 imp Exp $");

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <machine/bus.h>

#include <dev/uart/uart.h>
#include <dev/uart/uart_cpu.h>
#include <dev/uart/uart_bus.h>

#include <arm/ep93xx/ep93xxreg.h>

#include "uart_if.h"

/*
 * Clear pending interrupts.
 */
static void
epuart_clrint(struct uart_bas *bas)
{

	uart_setreg(bas, EP93XX_UART_Ctrl, 0);
}

static int
epuart_delay(struct uart_bas *bas)
{
	int divisor;

	uart_barrier(bas);
	divisor = uart_getreg(bas, EP93XX_UART_LinCL);
	divisor |= uart_getreg(bas, EP93XX_UART_LinCM) << 8;
	uart_barrier(bas);

	/* 1/10th the time to transmit 1 character (estimate). */
	/* XXX - what should this be? */
	return (16000000 * divisor / bas->rclk);
}

static int
epuart_divisor(int rclk, int baudrate)
{
	int actual_baud, divisor;
	int error;

	if (baudrate == 0)
		return (0);

	divisor = rclk / (16 * baudrate) - 1;
	if (divisor == 0 || divisor >= 65536)
		return (0);
	actual_baud = rclk / (16 * (divisor + 1));

	/* 10 times error in percent: */
	error = ((actual_baud - baudrate) * 2000 / baudrate + 1) >> 1;

	/* 3.0% maximum error tolerance: */
	if (error < -30 || error > 30)
		return (0);

	return (divisor);
}

static int
epuart_drain(struct uart_bas *bas, int what)
{
	int delay, limit;

	delay = epuart_delay(bas);

	if (what & UART_DRAIN_TRANSMITTER) {
		/*
		 * Pick an arbitrary high limit to avoid getting stuck in
		 * an infinite loop when the hardware is broken. Make the
		 * limit high enough to handle large FIFOs.
		 */
		limit = 10*1024;
		while ((uart_getreg(bas, EP93XX_UART_Flag) & EP93XX_UART_F_BUSY)
		    && --limit)
			DELAY(delay);
		if (limit == 0) {
			/* printf("epuart: transmitter appears stuck... "); */
			return (EIO);
		}
	}

	if (what & UART_DRAIN_RECEIVER) {
		/*
		 * Pick an arbitrary high limit to avoid getting stuck in
		 * an infinite loop when the hardware is broken. Make the
		 * limit high enough to handle large FIFOs and integrated
		 * UARTs. The HP rx2600 for example has 3 UARTs on the
		 * management board that tend to get a lot of data send
		 * to it when the UART is first activated.
		 */
		limit=10*4096;
		while (!(uart_getreg(bas, EP93XX_UART_Flag) &
		    EP93XX_UART_F_RXFE) && --limit) {
			(void)uart_getreg(bas, EP93XX_UART_Data);
			uart_barrier(bas);
			DELAY(delay << 2);
		}
		if (limit == 0) {
			/* printf("epuart: receiver appears broken... "); */
			return (EIO);
		}
	}

	return (0);
}

/*
 * The EP93xx com can't flush, just map them to drain.
 *
 * We can only flush UARTs with FIFOs. UARTs without FIFOs should be
 * drained. WARNING: this function clobbers the FIFO setting!
 */
static int
epuart_flush(struct uart_bas *bas, int what)
{

	return epuart_drain(bas, (what & UART_FLUSH_TRANSMITTER ?
	    UART_DRAIN_TRANSMITTER : 0) | (what & UART_FLUSH_RECEIVER ?
	    UART_DRAIN_RECEIVER : 0));
}

static int
epuart_param(struct uart_bas *bas, int baudrate, int databits, int stopbits,
    int parity)
{
	int divisor;
	uint8_t linch;

	linch = 0;
	if (databits >= 8)
		linch |= EP93XX_UART_8bit;
	else if (databits == 7)
		linch |= EP93XX_UART_7bit;
	else if (databits == 6)
		linch |= EP93XX_UART_6bit;
	else
		linch |= EP93XX_UART_5bit;
	if (stopbits > 1)
		linch |= EP93XX_UART_STP2;
	linch |= parity & 0x3;			/* no mark/space support? */
	linch |= EP93XX_UART_FIFOEn;		/* always enable FIFO */

	/* Set baudrate. */
	if (baudrate > 0) {
		divisor = epuart_divisor(bas->rclk, baudrate);
		if (divisor == 0)
			return (EINVAL);
		uart_barrier(bas);
		uart_setreg(bas, EP93XX_UART_LinCM, (divisor >> 8) & 0xff);
		uart_setreg(bas, EP93XX_UART_LinCL, divisor & 0xff);
	}

	uart_setreg(bas, EP93XX_UART_LinCH, linch);
	uart_barrier(bas);
	return (0);
}

/*
 * Low-level UART interface.
 */
static int epuart_probe(struct uart_bas *bas);
static void epuart_init(struct uart_bas *bas, int, int, int, int);
static void epuart_term(struct uart_bas *bas);
static void epuart_putc(struct uart_bas *bas, int);
static int epuart_poll(struct uart_bas *bas);
static int epuart_getc(struct uart_bas *bas);

struct uart_ops uart_epuart_ops = {
	.probe = epuart_probe,
	.init = epuart_init,
	.term = epuart_term,
	.putc = epuart_putc,
	.poll = epuart_poll,
	.getc = epuart_getc,
};

static int
epuart_probe(struct uart_bas *bas)
{

	if (bas->chan < 1 || bas->chan > 2)
		return (ENXIO);
	return (0);
}

static void
epuart_init(struct uart_bas *bas, int baudrate, int databits, int stopbits,
    int parity)
{

	/* Do the SYSCON stuff that NetBSD does? */
	if (bas->rclk == 0)
		bas->rclk = EP93XX_UART_FREQ;
	epuart_param(bas, baudrate, databits, stopbits, parity);

#if 0
	/* Set RTS & DTR. */
	uart_setreg(bas, EP93XX_UART_MdmCtrl, 0);
	uart_barrier(bas);
#endif
}

static void
epuart_term(struct uart_bas *bas)
{

	/* Clear RTS & DTR. */
	uart_setreg(bas, EP93XX_UART_MdmCtrl, 0);
	uart_barrier(bas);
}

static void
epuart_putc(struct uart_bas *bas, int c)
{
	int delay, limit;

	/* 1/10th the time to transmit 1 character (estimate). */
	delay = epuart_delay(bas);

	limit = 20;
	while ((uart_getreg(bas, EP93XX_UART_Flag) & EP93XX_UART_F_TXFF) &&
	    --limit)
		DELAY(delay);
	uart_setreg(bas, EP93XX_UART_Data, c);
	uart_barrier(bas);
}

/*
 * Handle RxSts in _poll and _getc?
 */
static int
epuart_poll(struct uart_bas *bas)
{

	if (!(uart_getreg(bas, EP93XX_UART_Flag) & EP93XX_UART_F_RXFE))
		return (uart_getreg(bas, EP93XX_UART_Data));
	return (-1);
}

static int
epuart_getc(struct uart_bas *bas)
{
	int delay;

	/* 1/10th the time to transmit 1 character (estimate). */
	delay = epuart_delay(bas);

	while ((uart_getreg(bas, EP93XX_UART_Flag) & EP93XX_UART_F_RXFE))
		DELAY(delay);
	return (uart_getreg(bas, EP93XX_UART_Data));
}

/*
 * High-level UART interface.
 */
struct epuart_softc {
	struct uart_softc base;
	u_int32_t	ctrl;
};

static int epuart_bus_attach(struct uart_softc *);
static int epuart_bus_detach(struct uart_softc *);
static int epuart_bus_flush(struct uart_softc *, int);
static int epuart_bus_getsig(struct uart_softc *);
static int epuart_bus_ioctl(struct uart_softc *, int, intptr_t);
static int epuart_bus_ipend(struct uart_softc *);
static int epuart_bus_param(struct uart_softc *, int, int, int, int);
static int epuart_bus_probe(struct uart_softc *);
static int epuart_bus_receive(struct uart_softc *);
static int epuart_bus_setsig(struct uart_softc *, int);
static int epuart_bus_transmit(struct uart_softc *);

static kobj_method_t epuart_methods[] = {
	KOBJMETHOD(uart_attach,		epuart_bus_attach),
	KOBJMETHOD(uart_detach,		epuart_bus_detach),
	KOBJMETHOD(uart_flush,		epuart_bus_flush),
	KOBJMETHOD(uart_getsig,		epuart_bus_getsig),
	KOBJMETHOD(uart_ioctl,		epuart_bus_ioctl),
	KOBJMETHOD(uart_ipend,		epuart_bus_ipend),
	KOBJMETHOD(uart_param,		epuart_bus_param),
	KOBJMETHOD(uart_probe,		epuart_bus_probe),
	KOBJMETHOD(uart_receive,	epuart_bus_receive),
	KOBJMETHOD(uart_setsig,		epuart_bus_setsig),
	KOBJMETHOD(uart_transmit,	epuart_bus_transmit),
	{ 0, 0 }
};

struct uart_class uart_epuart_class = {
	"epuart class",
	epuart_methods,
	sizeof(struct epuart_softc),
	.uc_range = 0x10000,
	.uc_rclk = EP93XX_UART_FREQ
};

#define	SIGCHG(c, i, s, d)				\
	if (c) {					\
		i |= (i & s) ? s : s | d;		\
	} else {					\
		i = (i & s) ? (i & ~s) | d : i;		\
	}

static int
epuart_bus_attach(struct uart_softc *sc)
{
	struct epuart_softc *epsc;
	struct uart_bas *bas;
	int mdmctrl;

	epsc = (struct epuart_softc *)sc;
	bas = &sc->sc_bas;

	epuart_bus_flush(sc, UART_FLUSH_RECEIVER|UART_FLUSH_TRANSMITTER);

	/* don't futz around with the ctrl reg if it's a console */
	if (bas->chan == 1) {
		mdmctrl = uart_getreg(bas, EP93XX_UART_MdmCtrl);
		if (mdmctrl & EP93XX_UART_DTR)
			sc->sc_hwsig |= SER_DTR;
		if (mdmctrl & EP93XX_UART_RTS)
			sc->sc_hwsig |= SER_RTS;
		epuart_bus_getsig(sc);
	}

	epuart_clrint(bas);

	epsc->ctrl = EP93XX_UART_MSIE | EP93XX_UART_RTIE |
	    EP93XX_UART_RIE | EP93XX_UART_UARTE;
	uart_setreg(bas, EP93XX_UART_Ctrl, epsc->ctrl);
	uart_barrier(bas);

	return (0);
}

static int
epuart_bus_detach(struct uart_softc *sc)
{
	struct epuart_softc *epsc;
	struct uart_bas *bas;

	epsc = (struct epuart_softc *)sc;
	bas = &sc->sc_bas;
	epuart_clrint(bas);

	/* disable UART and Ints */
	epsc->ctrl = 0;
	uart_setreg(bas, EP93XX_UART_Ctrl, epsc->ctrl);
	uart_barrier(bas);
	return (0);
}

static int
epuart_bus_flush(struct uart_softc *sc, int what)
{
	struct uart_bas *bas;
	int error;

	bas = &sc->sc_bas;
	mtx_lock_spin(&sc->sc_hwmtx);
	error = epuart_flush(bas, what);
	mtx_unlock_spin(&sc->sc_hwmtx);
	return (error);
}

static int
epuart_bus_getsig(struct uart_softc *sc)
{
	uint32_t new, sig;
	uint8_t mdmsts;

	/*
	 * We don't spin since we have already destroyed state, should we
	 * lock the hwsig w/ the spinlock?
	 */
	mtx_lock_spin(&sc->sc_hwmtx);
	mdmsts = uart_getreg(&sc->sc_bas, EP93XX_UART_MdmSts);
	sig = 0;
	if (mdmsts & EP93XX_UART_MS_DCD)
		sig |= SER_DCD;
	if (mdmsts & EP93XX_UART_MS_DDCD)
		sig |= SER_DDCD;
	if (mdmsts & EP93XX_UART_MS_RI)
		sig |= SER_RI;
	if (mdmsts & EP93XX_UART_MS_TERI)
		sig |= SER_DRI;
	if (mdmsts & EP93XX_UART_MS_DSR)
		sig |= SER_DSR;
	if (mdmsts & EP93XX_UART_MS_DDSR)
		sig |= SER_DDSR;
	if (mdmsts & EP93XX_UART_MS_CTS)
		sig |= SER_CTS;
	if (mdmsts & EP93XX_UART_MS_DCTS)
		sig |= SER_DCTS;
	new = sig & ~UART_SIGMASK_DELTA;
	sc->sc_hwsig = new;
	mtx_unlock_spin(&sc->sc_hwmtx);
	return (sig);
}

static int
epuart_bus_ioctl(struct uart_softc *sc, int request, intptr_t data)
{
	struct uart_bas *bas;
	int baudrate, divisor, error;
	uint8_t mdmctrl, lctrlh;

	bas = &sc->sc_bas;
	error = 0;
	mtx_lock_spin(&sc->sc_hwmtx);
	switch (request) {
	case UART_IOCTL_BREAK:
		lctrlh = uart_getreg(bas, EP93XX_UART_LinCH);
		if (data)
			lctrlh |= EP93XX_UART_BRK;
		else
			lctrlh &= ~EP93XX_UART_BRK;
		uart_setreg(bas, EP93XX_UART_LinCH, lctrlh);
		uart_barrier(bas);
		break;
	case UART_IOCTL_IFLOW:
		if (bas->chan == 1) {
			mdmctrl = uart_getreg(bas, EP93XX_UART_MdmCtrl);
			if (data)
				mdmctrl |= EP93XX_UART_RTS;
			else
				mdmctrl &= ~EP93XX_UART_RTS;
			uart_setreg(bas, EP93XX_UART_MdmCtrl, mdmctrl);
			uart_barrier(bas);
		} else
			error = ENODEV;
		break;
	case UART_IOCTL_OFLOW:
		/* hun? I can only read CTS */
		error = EINVAL;
		break;
	case UART_IOCTL_BAUD:
		divisor = uart_getreg(bas, EP93XX_UART_LinCL);
		divisor |= uart_getreg(bas, EP93XX_UART_LinCM) << 8;
		uart_barrier(bas);
		baudrate = bas->rclk / (16 * (divisor + 1));
		if (baudrate > 0)
			*(int*)data = baudrate;
		else
			error = ENXIO;
		break;
	default:
		error = EINVAL;
		break;
	}
	mtx_unlock_spin(&sc->sc_hwmtx);
	return (error);
}

static int
epuart_bus_ipend(struct uart_softc *sc)
{
	struct epuart_softc *epsc;
	struct uart_bas *bas;
	int ipend;
	uint8_t iidclr, rxsts;

	epsc = (struct epuart_softc *)sc;
	bas = &sc->sc_bas;
#if 0
	iidclr = uart_getreg(bas, EP93XX_UART_IntIDCLR);
	printf("%c%c", (int)('0' + iidclr), (int)('0' +
	    uart_getreg(bas, EP93XX_UART_RXSts)));
	mtx_lock_spin(&sc->sc_hwmtx);
#else
	mtx_lock_spin(&sc->sc_hwmtx);
	iidclr = uart_getreg(bas, EP93XX_UART_IntIDCLR);
#endif
	if (!(iidclr & (EP93XX_UART_TIS | EP93XX_UART_RIS |
	    EP93XX_UART_RTIS | EP93XX_UART_MIS))) {
		mtx_unlock_spin(&sc->sc_hwmtx);
		return (0);
	}
	ipend = 0;
	if ((iidclr & EP93XX_UART_RIS) || (iidclr & EP93XX_UART_RTIS)) {
		rxsts = uart_getreg(bas, EP93XX_UART_RXSts);
		if (rxsts & EP93XX_UART_OE) {
			ipend |= UART_IPEND_OVERRUN;
			uart_setreg(bas, EP93XX_UART_RXSts, 0);	/* clear */
		}
		if (rxsts & EP93XX_UART_BE)
			ipend |= UART_IPEND_BREAK;
		ipend |= UART_IPEND_RXREADY;
		epsc->ctrl &= ~(EP93XX_UART_RIE | EP93XX_UART_RTIE);
		uart_setreg(bas, EP93XX_UART_Ctrl, epsc->ctrl);
	}
	if (iidclr & EP93XX_UART_TIS) {
		ipend |= UART_IPEND_TXIDLE;
		epsc->ctrl &= ~EP93XX_UART_TIE;
		uart_setreg(bas, EP93XX_UART_Ctrl, epsc->ctrl);
	}
	if (iidclr & EP93XX_UART_MIS)
		ipend |= UART_IPEND_SIGCHG;

	uart_setreg(bas, EP93XX_UART_IntIDCLR, 0);	/* clear */
	mtx_unlock_spin(&sc->sc_hwmtx);
	return ((sc->sc_leaving) ? 0 : ipend);
}

static int
epuart_bus_param(struct uart_softc *sc, int baudrate, int databits,
    int stopbits, int parity)
{
	struct uart_bas *bas;
	int error;

	bas = &sc->sc_bas;
	mtx_lock_spin(&sc->sc_hwmtx);
	error = epuart_param(bas, baudrate, databits, stopbits, parity);
	mtx_unlock_spin(&sc->sc_hwmtx);
	return (error);
}

static int
epuart_bus_probe(struct uart_softc *sc)
{
	struct uart_bas *bas;
	int error;

	bas = &sc->sc_bas;

	error = epuart_probe(bas);
	if (error)
		return (error);

	if (sc->sc_sysdev == NULL) {
		/* By using epuart_init() we also set DTR and RTS. */
		epuart_init(bas, 9600, 8, 1, UART_PARITY_NONE);
	}

	error = epuart_drain(bas, UART_DRAIN_TRANSMITTER);
	if (error)
		return (error);

	/*
	 * XXX - Set loopback mode.  I guess we should check operation via
	 * loopback mode.
	 */

	if (bas->chan == 1)
		device_set_desc(sc->sc_dev,
		    "UART1 with HDLC and Modem Control Signals");
	else
		device_set_desc(sc->sc_dev,
		    "UART2 with IrDA");

	sc->sc_txfifosz = 16;
#if 0
	/*
	 * XXX there are some issues related to hardware flow control and
	 * it's likely that uart(4) is the cause. This basicly needs more
	 * investigation, but we avoid using for hardware flow control
	 * until then.
	 */
#endif

	return (0);
}

static int
epuart_bus_receive(struct uart_softc *sc)
{
	struct epuart_softc *epsc;
	struct uart_bas *bas;
	int xc;
	uint8_t flag, rxsts;

	epsc = (struct epuart_softc *)sc;
	bas = &sc->sc_bas;
	mtx_lock_spin(&sc->sc_hwmtx);
	while (!((flag = uart_getreg(bas, EP93XX_UART_Flag)) &
	    EP93XX_UART_F_RXFE)) {
		if (uart_rx_full(sc)) {
			sc->sc_rxbuf[sc->sc_rxput] = UART_STAT_OVERRUN;
			break;
		}
		xc = uart_getreg(bas, EP93XX_UART_Data);
		printf("%c", (int)xc);
		rxsts = uart_getreg(bas, EP93XX_UART_RXSts);
		if (rxsts & EP93XX_UART_BE)
			xc |= UART_STAT_BREAK;
		if (rxsts & EP93XX_UART_FE)
			xc |= UART_STAT_FRAMERR;
		if (rxsts & EP93XX_UART_PE)
			xc |= UART_STAT_PARERR;
		uart_rx_put(sc, xc);
	}
	/* Discard everything left in the Rx FIFO. */
	while (flag & EP93XX_UART_F_RXFE) {
		(void)uart_getreg(bas, EP93XX_UART_Data);
		uart_barrier(bas);
		flag = uart_getreg(bas, EP93XX_UART_Flag);
	}
	/* reenable RX interrupts */
	epsc->ctrl |= EP93XX_UART_RIE | EP93XX_UART_RTIE;
	uart_setreg(bas, EP93XX_UART_Ctrl, epsc->ctrl);
	mtx_unlock_spin(&sc->sc_hwmtx);
 	return (0);
}

static int
epuart_bus_setsig(struct uart_softc *sc, int sig)
{
	struct uart_bas *bas;
	uint32_t new, old;
	int mdmctrl;

	bas = &sc->sc_bas;
	do {
		old = sc->sc_hwsig;
		new = old;
		if (sig & SER_DDTR) {
			SIGCHG(sig & SER_DTR, new, SER_DTR,
			    SER_DDTR);
		}
		if (sig & SER_DRTS) {
			SIGCHG(sig & SER_RTS, new, SER_RTS,
			    SER_DRTS);
		}
	} while (!atomic_cmpset_32(&sc->sc_hwsig, old, new));
	mtx_lock_spin(&sc->sc_hwmtx);
	mdmctrl = 0;
	if (new & SER_DTR)
		mdmctrl |= EP93XX_UART_DTR;
	if (new & SER_RTS)
		mdmctrl |= EP93XX_UART_RTS;
	uart_setreg(bas, EP93XX_UART_MdmCtrl, mdmctrl);
	uart_barrier(bas);
	mtx_unlock_spin(&sc->sc_hwmtx);
	return (0);
}

static int
epuart_bus_transmit(struct uart_softc *sc)
{
	struct epuart_softc *epsc;
	struct uart_bas *bas;
	int i;

	epsc = (struct epuart_softc *)sc;
	bas = &sc->sc_bas;
	mtx_lock_spin(&sc->sc_hwmtx);
	/* XXX - ugly, this sucks, but that's the way uart forces us to be */
	while ((uart_getreg(bas, EP93XX_UART_Flag) & EP93XX_UART_F_TXFE) == 0)
		;
	epsc->ctrl |= EP93XX_UART_TIE;
	uart_setreg(bas, EP93XX_UART_Ctrl, epsc->ctrl);
	uart_barrier(bas);
	for (i = 0; i < sc->sc_txdatasz; i++) {
		uart_setreg(bas, EP93XX_UART_Data, sc->sc_txbuf[i]);
		uart_barrier(bas);
	}
	sc->sc_txbusy = 1;
	mtx_unlock_spin(&sc->sc_hwmtx);
	return (0);
}
